from ..imports import *

__all__ = ["MultiRainbow", "compare_rainbows"]


class MultiRainbow:
    """
    MultiRainbow objects collect multiple Rainbow objects together,
    providing a quick interface to apply the same action or visualization
    to all of them at once. It's meant to be a tool to facilitate
    quick comparisons between different pipeline analyses of the
    same dataset.
    """

    def __init__(self, rainbows, names=None):
        """
        Initialize from a list of Rainbows.

        Parameters
        ----------
        rainbows : list
            A list containing two or more Rainbow objects.
        names : list
            A list of names for the Rainbows. These will overwrite
            the names provided in the `.metadata['name']` entry
            of each input Rainbow.
        """

        if not isinstance(rainbows, list):
            raise RuntimeError("Please provide a list of 🌈s.")
        if len(rainbows) < 2:
            raise RuntimeError("Please provide more than one 🌈.")

        # decouple these rainbow from their inputs
        self.rainbows = [x._create_copy() for x in rainbows]

        # keep track of the names
        self.names = names or [(x.name or i) for i, x in enumerate(rainbows)]
        if len(np.unique(self.names)) < len(self.names):
            message = """
            The input names {self.names}
            are not unique. Proceed with caution, or provide
            each 🌈 with a unique name. You can do so either by
            giving each individual 🌈 a name with `rainbow.name='?!?!?!'`
            or by supplying a list of unique names to the `names=`
            keyword argument when starting your comparison.
            """
            cheerfully_suggest(message)

        # make sure the names and rainbows match up
        assert len(self.names) == len(self.rainbows)

        # pull units from first rainbow
        self.w_unit = u.Unit(self.list_of_rainbows[0].wavelength.unit)
        self.t_unit = u.Unit(self.list_of_rainbows[0].time.unit)

    @property
    def list_of_rainbows(self):
        return self.rainbows

    @property
    def dict_of_rainbows(self):
        return {k: v for k, v in zip(self.names, self.rainbows)}

    def __repr__(self):
        """
        How should this object be represented?
        """
        return f"<MultiRainbow({self.rainbows})>"

    @property
    def nrainbows(self):
        """
        What's the number of rainbows in here?
        """
        return len(self.rainbows)

    def _setup_panels(self, rows=1, figsize=None, **kw):
        """
        Set up a grid of panels to plot into.

        Parameters
        ----------
        rows : int
            The number of rows into which to arrange the
            plot panels. (This will probably only be needed
            for working with more than about 3 Rainbows.)
        figsize : tuple
            The size (width, height) for the figure to create.
            If left blank, this will be estimated from the
            default matplotlib figsize and the number of
            figures being produced.
        """

        # estimate a figure size from current matplotlib defaults
        if figsize == None:
            default_figsize = plt.matplotlib.rcParams["figure.figsize"]
            figsize = [self.nrainbows * default_figsize[0], default_figsize[1]]

        # create the figure and grid of axes as subplots
        self.figure, self.axes = plt.subplots(
            rows,
            int(np.ceil(self.nrainbows / rows)),
            sharex=True,
            sharey=True,
            figsize=figsize,
            constrained_layout=True,
        )

        # give titles to the plot panels
        if self.names is not None:
            for name, ax in zip(self.names, self.axes):
                ax.set_title(name)

    def _guess_good_uniform_wavelength_grid(
        self, fraction_to_supersample=0.5, wscale=None, plot=False
    ):
        """
        Guess a good shared wavelength grid that would work
        reasonably well for all the Rainbows in this set.

        Parameters
        ----------
        fraction_to_supersample : float
            When setting the new wavelength bins, what fraction of bins
            in the original Rainbows will be super-sampled? 0.0 will create
            new bins that are just larger than the largest original bin,
            1.0 will create new bins that are just smaller than the smallest
            original bin, and 0.5 is the default.

        plot : bool
            Should we make a plot of the d[wavelength]/d[bin], for diagnostics?
        """

        # create a list of interpolating functions for the datasets' dw
        dw_interpolators = [
            interp1d(
                r.wavelength.to_value(self.w_unit),
                np.gradient(r.wavelength.to_value(self.w_unit)),
                fill_value=np.inf,
                bounds_error=False,
            )
            for r in self.list_of_rainbows
        ]

        def smallest_dw(w):
            """
            Helper function to determine the smallest dw
            in the datasets at any particular wavelength.

            Parameters
            ----------
            w : np.array
                Wavelength (with no units).
            """

            dw = np.inf * np.ones_like(w)
            for interpolator in dw_interpolators:
                dw = np.minimum(dw, interpolator(w))
            return dw

        # compile w and dw, weighted by how many appear in each dataset
        w_represented = np.sort(
            np.hstack(
                [r.wavelength.to_value(self.w_unit) for r in self.list_of_rainbows]
            )
        )
        dw_represented = smallest_dw(w_represented)

        # fit a line to log(dw) as a function of log(w)
        slope, intercept = np.polyfit(np.log(w_represented), np.log(dw_represented), 1)
        dw_model = np.exp(np.polyval([slope, intercept], np.log(w_represented)))

        # decide if it should be linear or logarithmic
        if wscale is None:
            if slope < 0.5:
                wscale = "linear"
            else:
                wscale = "log"

        if wscale == "linear":
            dw = np.percentile(dw_represented, 100 * (1 - fraction_to_supersample))
            new_wavelengths = np.arange(
                np.nanmin(w_represented), np.nanmax(w_represented), dw
            )
        elif wscale == "log":
            R = np.percentile(
                w_represented / dw_represented, 100 * fraction_to_supersample
            )
            new_wavelengths = np.exp(
                np.arange(
                    np.log(np.nanmin(w_represented)),
                    np.log(np.nanmax(w_represented)),
                    1 / R,
                )
            )
        else:
            raise RuntimeError('🌈 Currently only "log" or "linear" are supported.')

        if plot:

            # plot the individual datasets
            for r in self.list_of_rainbows:
                plt.plot(
                    r.wavelength, np.gradient(r.wavelength), marker="o", alpha=0.25
                )

            # plot the minimum dw at each wavelength
            plt.plot(
                w_represented, dw_represented, marker="o", color="black", alpha=0.25
            )

            # convert to log scale, so R=constant is a slope=1 line
            plt.yscale("log")
            plt.xscale("log")
            plt.axis("scaled")

            # plot the line fit result
            plt.plot(
                w_represented,
                dw_model,
                linestyle="--",
                color="black",
                linewidth=2,
                alpha=0.25,
            )

            # plot the new logarithmic or linear wavelength grid
            plt.plot(
                new_wavelengths,
                np.gradient(new_wavelengths),
                color="black",
                linewidth=20,
                alpha=0.25,
            )
            plt.title(f"{slope:.3} -> {wscale}!")
            plt.ylabel("d[wavelength]/d[bin]")

        return new_wavelengths * self.w_unit

    def _check_if_wavelengths_are_aligned(self):
        """
        Check whether the wavelength grid is already aligned.

        Returns
        -------
        aligned : bool
            Are the wavelengths the same across all?
        """

        first_rainbow = self.list_of_rainbows[0]

        # check if they're already aligned
        wavelengths_are_same_size = np.all(
            [np.all(r.nwave == first_rainbow.nwave) for r in self.list_of_rainbows]
        )

        if wavelengths_are_same_size:
            wavelengths_are_already_aligned = np.all(
                [
                    np.all(r.wavelength == first_rainbow.wavelength)
                    for r in self.list_of_rainbows
                ]
            )
        else:
            wavelengths_are_already_aligned = False

        return wavelengths_are_already_aligned

    def _check_if_times_are_aligned(self):
        """
        Check whether the time grid is already aligned.

        Returns
        -------
        aligned : bool
            Are the times the same across all Rainbows?
        """
        # check if they're already aligned
        times_are_same_size = np.all(
            [
                np.all(r.ntime == self.list_of_rainbows[0].ntime)
                for r in self.list_of_rainbows
            ]
        )

        if times_are_same_size:
            times_are_already_aligned = np.all(
                [
                    np.all(r.time == self.list_of_rainbows[0].wavelength)
                    for r in self.list_of_rainbows
                ]
            )
        else:
            times_are_already_aligned = False

        return times_are_already_aligned

    @property
    def wavelength(self):
        """
        The shared wavelength axis.
        """
        if self._check_if_wavelengths_are_aligned():
            return self.list_of_rainbows[0].wavelength
        else:
            raise RuntimeError(f"🌈 {self} has more than one wavelength array.")

    @property
    def time(self):
        """
        The shared time axis.
        """
        if self._check_if_times_are_aligned():
            return self.list_of_rainbows[0].time
        else:
            raise RuntimeError(f"🌈 {self} has more than one time array.")

    def align_wavelengths(self, **kw):
        """
        Make a new MultiRainbow where wavelengths have been aligned
        across all the Rainbows, making a guess for a reasonable
        new wavelength grid.

        Parameters
        ----------
        kw : dict
            Extra keyword arguments will be passed to
            `._guess_good_uniform_wavelength_grid`
        """

        # if the wavelengths are already aligned, don't do anything!
        if self._check_if_wavelengths_are_aligned():
            return self
        else:
            # come up with a decent guess for a shared wavelength grid
            w = self._guess_good_uniform_wavelength_grid(**kw)

            # bin to that shared grid, and don't trim nans at the edges
            return self.bin(wavelength=w, trim=False)

    def __getitem__(self, key):
        """
        Trim a rainbow by indexing, slicing, or masking.
        Two indices must be provided (`[:,:]`).

        Examples
        --------
        ```
        r[:,:]
        r[10:20, :]
        r[np.arange(10,20), :]
        r[r.wavelength > 1*u.micron, :]
        r[:, np.abs(r.time) < 1*u.hour]
        r[r.wavelength > 1*u.micron, np.abs(r.time) < 1*u.hour]
        ```

        Parameters
        ----------
        key : tuple
            The (wavelength, time) slices, indices, or masks.
        """
        new_rainbows = [r.__getitem__(key) for r in self.list_of_rainbows]
        return MultiRainbow(new_rainbows, names=self.names)

    def normalize(self, **kwargs):
        """
        Normalize by dividing through by the median spectrum and/or lightcurve.

        (see `Rainbow.normalize` for input options)

        Returns
        -------
        normalized : MultiRainbow
            The normalized MultiRainbow.
        """
        new_rainbows = [r.normalize(**kwargs) for r in self.list_of_rainbows]
        return MultiRainbow(new_rainbows, names=self.names)

    def bin(self, **kwargs):
        """
        Bin the rainbow in wavelength and/or time.

        (see `Rainbow.bin` for input options)

        Returns
        -------
        binned : MultiRainbow
            The binned MultiRainbow.
        """
        new_rainbows = [r.bin(**kwargs) for r in self.list_of_rainbows]
        return MultiRainbow(new_rainbows, names=self.names)

    def plot(self, **kwargs):
        """
        Plot flux as sequence of offset light curves.

        (see `Rainbow.plot` for input options)
        """

        # set up a grid of panels
        self._setup_panels()

        # make all the individual plots
        for r, a in zip(self.list_of_rainbows, self.axes):
            r.plot(ax=a, **kwargs)

    def imshow(
        self, vmin=None, vmax=None, rows=1, figsize=None, colorbar=True, **kwargs
    ):
        """
        imshow flux as a function of time (x = time, y = wavelength, color = flux).

        Parameters
        ----------
        vmin : float
            The bottom of the color scale.
        vmax : float
            The top of the color scale.
        rows : int
            Over how many rows should the panels be distributed?
        figsize : tuple
            The (width, height) of the total figure.
        colorbar : bool
            Should we include a colorbar?
        quantity : str
            The fluxlike quantity to imshow.
            (Must be a key of `rainbow.fluxlike`).
        w_unit : str, astropy.unit.Unit
            The unit for plotting wavelengths.
        t_unit : str, astropy.unit.Unit
            The unit for plotting times.
        aspect : str
            What aspect ratio should be used for the imshow?
        kwargs : dict
            Most other keyword arguments will be passed on into
            `plt.imshow`. If there's an `plt.imshow` keyword you
            want to set, try it!
        """

        # set up a grid of panels
        self._setup_panels(rows=rows, figsize=figsize)

        # figure out a good shared color limits (unless already supplied)
        vmin = vmin or np.nanmin(
            [np.nanmin(u.Quantity(r.flux).value) for r in self.list_of_rainbows]
        )
        vmax = vmax or np.nanmax(
            [np.nanmax(u.Quantity(r.flux).value) for r in self.list_of_rainbows]
        )

        # make all the individual imshows
        for r, a, i in zip(self.list_of_rainbows, self.axes, range(self.nrainbows)):
            r.imshow(ax=a, vmin=vmin, vmax=vmax, colorbar=False, **kwargs)
            if i == 0:
                xlabel, ylabel = a.get_xlabel(), a.get_ylabel()
            a.set_xlabel("")
            a.set_ylabel("")

        self.figure.supxlabel(xlabel)
        self.figure.supylabel(ylabel)

        # add a colorbar, but steal from all the axes
        if colorbar:
            plt.colorbar(ax=self.axes)

        # set the ylimits to span everything
        plt.ylim(
            np.nanmax([x._imshow_extent[2] for x in self.list_of_rainbows]),
            np.nanmin([x._imshow_extent[3] for x in self.list_of_rainbows]),
        )

    def animate_lightcurves(
        self,
        filename="multi-animated-lightcurves.gif",
        fps=None,
        dpi=None,
        **kwargs,
    ):
        """
        Create an animation to show how the lightcurve changes
        as we flip through every wavelength.

        Parameters
        ----------
        filename : str
            Name of file you'd like to save results in.
            Currently supports only .gif files.
        fps : float
            frames/second of animation
        xlim : tuple
            Custom xlimits for the plot
        ylim : tuple
            Custom ylimits for the plot
        cmap : str, matplotlib.colors.Colormap
            The color map to use for expressing wavelength
        vmin : astropy.units.Quantity
            The minimum value to use for the wavelength colormap
        vmax : astropy.units.Quantity
            The maximum value to use for the wavelength colormap
        scatterkw : dict
            A dictionary of keywords to be passed to `plt.scatter`
        textkw : dict
            A dictionary of keywords to be passed to `plt.text`
        """

        # make sure there's the same number of wavelengths for all rainbows
        assert np.all(
            [r.nwave == self.list_of_rainbows[0].nwave for r in self.list_of_rainbows]
        )

        # set up a grid of panels
        self._setup_panels()

        # setup all the individual plots
        for r, a in zip(self.list_of_rainbows, self.axes):
            r._setup_animate_lightcurves(ax=a, **kwargs)

        # the figure should be the same for all panels
        figure = r._animate_lightcurves_components["fi"]

        def update(frame):
            """
            This function will be called to update each frame
            of the animation.

            Parameters
            ----------
            frame : int
                An integer that will advance with each frame.
            """
            artists = []
            for r in self.list_of_rainbows:
                r._animate_lightcurves_components["update"](frame)
                for k in ["text", "scatter"]:
                    artists.append(r._animate_lightcurves_components[k])
            return artists

        # make and save the animation
        animator = ani.FuncAnimation(
            figure,
            update,
            frames=np.arange(0, r.nwave),
            blit=True,
        )
        animator.save(
            filename, fps=fps, dpi=dpi, savefig_kwargs=dict(facecolor="white")
        )
        plt.close()

    def animate_spectra(
        self, filename="multi-animated-spectra.gif", fps=None, dpi=None, **kwargs
    ):
        """
        Create an animation to show how the spectrum changes
        as we flip through every timepoint.

        Parameters
        ----------
        filename : str
            Name of file you'd like to save results in.
            Currently supports only .gif files.
        fps : float
            frames/second of animation
        xlim : tuple
            Custom xlimits for the plot
        ylim : tuple
            Custom ylimits for the plot
        cmap : str, matplotlib.colors.Colormap
            The color map to use for expressing wavelength
        vmin : astropy.units.Quantity
            The minimum value to use for the wavelength colormap
        vmax : astropy.units.Quantity
            The maximum value to use for the wavelength colormap
        scatterkw : dict
            A dictionary of keywords to be passed to `plt.scatter`
        textkw : dict
            A dictionary of keywords to be passed to `plt.text`
        """

        assert np.all(
            [r.ntime == self.list_of_rainbows[0].ntime for r in self.list_of_rainbows]
        )

        # set up a grid of panels
        self._setup_panels()

        # setup all the individual plots
        for r, a in zip(self.list_of_rainbows, self.axes):
            r._setup_animate_spectra(ax=a, **kwargs)

        # the figure should be the same for all panels
        figure = r._animate_spectra_components["fi"]

        def update(frame):
            """
            This function will be called to update each frame
            of the animation.

            Parameters
            ----------
            frame : int
                An integer that will advance with each frame.
            """
            artists = []
            for r in self.list_of_rainbows:
                r._animate_spectra_components["update"](frame)
                for k in ["text", "scatter"]:
                    artists.append(r._animate_spectra_components[k])
            return artists

        # make and save the animation
        animator = ani.FuncAnimation(
            figure,
            update,
            frames=np.arange(0, r.ntime),
            blit=True,
        )
        animator.save(
            filename, fps=fps, dpi=dpi, savefig_kwargs=dict(facecolor="white")
        )
        plt.close()


def compare_rainbows(*args, **kwargs):
    """
    Compare a group of Rainbows to each other.

    Parameters
    ----------
    rainbows : list
        A list containing two or more Rainbow objects.

    Returns
    -------
    multirainbow : MultiRainbow
        A MultiRainbow object that contains all of the rainbows to
        be compared. This object can do many of the same actions or
        visualizations as a normal Rainbow, but it will (attempt to)
        automatically apply them to every Rainbow.
    """
    return MultiRainbow(*args, **kwargs)



from ....imports import *

__all__ = ["get_median_spectrum"]


def get_median_spectrum(self):
    """
    Return a spectrum of the star, medianed over all times.

    Parameters
    ----------

    Returns
    -------
    median_spectrum : np.array (wavelike)
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return np.nanmedian(self.flux, axis=1)



# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["from_radica"]


def from_radica(self, filepath, order=1):
    """
    Populate a Rainbow from a file in the radica format.

    Parameters
    ----------

    self : Rainbow
        The object to be populated.
    filepath : str
        The path to the file to load.
    """

    hdu = fits.open(filepath)

    n_orders = 2
    cheerfully_suggest(
        f"""
    Loading NIRISS spectroscopic `order={order}``. Two orders are available,
    and you can set which (1,2) you want to read with the `order=` option.
    """
    )
    self.fluxlike["wavelength_2d"] = (
        np.transpose(hdu[f"Wave 2D Order {order}"].data * u.micron) * 1
    )
    self.wavelike["wavelength"] = (
        np.nanmedian(self.fluxlike["wavelength_2d"], axis=1) * 1
    )
    self.fluxlike["flux"] = (
        np.transpose(hdu[f"Flux Order {order}"].data * u.electron) * 1
    )
    self.fluxlike["uncertainty"] = (
        np.transpose(hdu[f"Flux Error Order {order}"].data * u.electron) * 1
    )
    astropy_times = Time(hdu["Time"].data, format="jd", scale="tdb")
    self.set_times_from_astropy(astropy_times, is_barycentric=True)



"""
Define a writer for chromatic .rainbow.npy files.
"""


# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["to_text"]


def to_text(self, filepath, overwrite=True, group_by="wavelength"):
    """
    Write a Rainbow to a file in the text format.

    Parameters
    ----------

    self : Rainbow
        The object to be saved.

    filepath : str
        The path to the file to write.
    """

    # a 1D array of wavelengths (with astropy units of length)
    the_1D_array_of_wavelengths = self.wavelike["wavelength"]

    # a 1D array of times (with astropy units of time)
    the_1D_array_of_times = self.timelike["time"]

    # a 2D (row = wavelength, col = array of fluxes
    the_2D_array_of_fluxes = self.fluxlike["flux"]

    # write out your file, however you like
    w, t = np.meshgrid(self.wavelength.to("micron"), self.time.to("day"), indexing="ij")

    def make_into_columns(x):
        if group_by == "wavelength":
            return x.flatten()
        if group_by == "time":
            return x.T.flatten()

    # create a table
    table = Table(
        dict(wavelength=make_into_columns(w), time=make_into_columns(t)),
        meta=self.metadata,
    )
    table["flux"] = make_into_columns(self.flux)
    try:
        table["uncertainty"] = make_into_columns(self.uncertainty)
    except KeyError:
        pass

    other_keys = list(self.fluxlike.keys())
    other_keys.remove("flux")
    other_keys.remove("uncertainty")

    for k in other_keys:
        table[k] = make_into_columns(self.fluxlike[k])

    table.write(filepath, format="ascii.ecsv", overwrite=overwrite)



"""
This module serves as a template for creating a new Rainbow
writer. If you want to add the ability to writer chromatic
light curves to a new kind of file format, a good process
would be to do something like the following

    1. Copy this `template.py` file into a new file in the
    `writers/` directory, ideally with a name that's easy
    to recognize, such as `writers/abcdefgh.py` (assuming
    `abcdefgh` is the name of your format)

    2. Start by finding and replacing `abcdefgh` in this
    template with the name of your format.

    3. Edit the `to_abcdefgh` function so that it will
    write out a Rainbow to whatever format you want.

    4. Edit the `writers/__init__.py` file to import your
    `to_abcdefgh` function to be accessible when people
    try to write Rainbows out to files. Add an `elif` statement
    to the `guess_writer` function that will help guess which
    writer to use from some aspect(s) of the filename.

    (This `guess_writer` function also accepts a `format=`
    keyword that allows the user to explicitly specify that
    the abcdefgh writer should be used.)

    5. Submit a pull request to the github repository for
    this package, so that other folks can use your handy
    new writer too!
"""

# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["to_abcdefgh"]


def to_abcdefgh(self, filepath):
    """
    Write a Rainbow to a file in the abcdefgh format.

    Parameters
    ----------

    self : Rainbow
        The object to be saved.

    filepath : str
        The path to the file to write.
    """

    # a 1D array of wavelengths (with astropy units of length)
    the_1D_array_of_wavelengths = self.wavelike["wavelength"]

    # a 1D array of times (with astropy units of time)
    the_1D_array_of_times = self.timelike["time"]

    # a 2D (row = wavelength, col = array of fluxes
    the_2D_array_of_fluxes = self.fluxlike["flux"]

    # write out your file, however you like
    write_to_abcdefgh(
        filepath,
        the_1D_array_of_wavelengths,
        the_1D_array_of_times,
        the_2D_array_of_fluxes,
        **some_other_stuff_too_maybe,
    )

    # add some warnings if there's any funny business
    if something_goes_wonky():
        cheerfully_suggest(
            f"""
        Here's a potential problem that the user should know about.
        """
        )



from ..rainbows import *
from .setup_tests import *


def test_lightcurve_and_spectrum_summaries():
    fi, ax = plt.subplots(1, 2, sharey=True)
    s = SimulatedRainbow().inject_noise().inject_transit()
    ax[0].plot(s.time, s.get_average_lightcurve())
    ax[1].plot(s.wavelength, s.get_average_spectrum())
    plt.savefig(
        os.path.join(
            test_directory, "demonstration-of-average-lightcurve-and-spectrum.pdf"
        )
    )


def test_measured_scatter_summary():
    # make a fake Rainbow with known noise
    signal_to_noise = 100
    s = SimulatedRainbow().inject_noise(signal_to_noise=signal_to_noise)

    # mathematically, what do we expect?
    expected_sigma = 1 / signal_to_noise
    uncertainty_on_expected_sigma = expected_sigma / np.sqrt(2 * (s.ntime - 1))
    # (see equation 3.48 of Sivia and Skilling for the uncertainty on sigma)

    # loop through methods
    methods = ["standard-deviation", "MAD"]
    fi, ax = plt.subplots(
        1,
        len(methods),
        sharey=True,
        sharex=True,
        figsize=(8, 3),
        dpi=300,
        constrained_layout=True,
        facecolor="white",
    )
    for i, method in enumerate(methods):
        # calculate the measured scatter
        measured_scatter = s.get_measured_scatter(method=method)

        # plot the measured scatters + 1-sigma expectation for its uncertainty
        plt.sca(ax[i])
        plt.plot(s.wavelength, measured_scatter, marker="o", color="black")
        plt.axhline(expected_sigma, color="black", alpha=0.3)
        plt.axhspan(
            expected_sigma - uncertainty_on_expected_sigma,
            expected_sigma + uncertainty_on_expected_sigma,
            color="gray",
            alpha=0.3,
        )

        # plot range that shouldn't be exceeded in this test
        many_sigma = 10
        plt.axhspan(
            expected_sigma - uncertainty_on_expected_sigma * many_sigma,
            expected_sigma + uncertainty_on_expected_sigma * many_sigma,
            color="red",
            alpha=0.05,
            zorder=-10,
        )
        plt.ylim(0, None)
        plt.xscale("log")
        plt.xlabel(f'Wavelength ({s.wavelength.unit.to_string("latex_inline")})')
        plt.ylabel(f"Measured Scatter\n('{method}')")

        # throw an error if the measured scatter gets too large
        assert np.all(
            np.abs(measured_scatter - expected_sigma)
            < uncertainty_on_expected_sigma * many_sigma
        )

    plt.savefig(os.path.join(test_directory, "demonstration-of-measured-scatter.pdf"))



from ...imports import *
from ...resampling import *

__all__ = ["_create_shared_wavelength_axis", "align_wavelengths"]


def _create_shared_wavelength_axis(
    rainbow, wscale="linear", supersampling=1, visualize=False
):
    """
    Create a shared 1D wavelength axis that tries to compress
    different wavelength axes associated with different times.

    Parameters
    ----------
    wscale : str
        What kind of a new wavelength axis should be created?
        Options include:
            'linear' = constant d[wavelength] between grid points
            'log' = constant d[wavelength]/[wavelength] between grid points

    supersampling : float
        By how many times should we increase or decrease the wavelength sampling?
        In general, values >1 will split each input wavelength grid point into
        multiple supersampled wavelength grid points, values close to 1 will
        produce approximately one output wavelength for each input wavelength,
        and values <1 will average multiple input wavelengths into a single output
        wavelength bin.

        Unless this is significantly less than 1, there's a good chance your output
        array may have strong correlations between one or more adjacent wavelengths.
        Be careful when trying to use the resulting uncertainties!

        (FIXME = add a way to estimate a covariance matrix when binning?)

    visualize : bool
        Should we make some plots showing how the shared wavelength
        axis compares to the original input wavelength axes?
    """
    w = rainbow.fluxlike["wavelength_2d"] * 1
    w[rainbow.ok == False] = np.nan
    dw_per_time = np.gradient(w, axis=rainbow.waveaxis)
    R_per_time = w / dw_per_time

    if visualize:
        fi, ax = plt.subplots(1, 2, figsize=(8, 3), dpi=300)
        plt.sca(ax[0])
        plt.imshow(dw_per_time, aspect="auto", vmin=0)
        plt.xlabel("Time Index")
        plt.ylabel("Wavelength Index")
        plt.title(r"$\Delta\lambda$")
        plt.colorbar(orientation="horizontal", pad=0.25)

        plt.sca(ax[1])
        plt.imshow(R_per_time, aspect="auto", vmin=0)
        plt.xlabel("Time Index")
        plt.ylabel("Wavelength Index")
        plt.title(r"R = $\lambda/\Delta\lambda$")
        plt.colorbar(orientation="horizontal", pad=0.25)

    min_w, max_w = np.nanmin(w).to("micron").value, np.nanmax(w).to("micron").value
    if wscale == "linear":
        dw = np.nanmedian(dw_per_time).to("micron").value / supersampling
        shared_w = np.arange(min_w, max_w + dw, dw) * u.micron
    elif wscale == "log":
        R = np.nanmedian(w / dw_per_time) * supersampling
        shared_w = np.exp(np.arange(np.log(min_w), np.log(max_w) + 1 / R, 1 / R))
    shared_dw = np.gradient(shared_w)
    shared_R = shared_w / shared_dw

    if visualize:
        fi, ax = plt.subplots(1, 2, figsize=(8, 3), dpi=300)
        plt.sca(ax[0])
        plt.plot(w, dw_per_time, alpha=0.2)
        plt.plot(
            shared_w,
            shared_dw * supersampling,
            color="black",
            label=f"{supersampling}x(shared)",
            marker=".",
        )
        plt.title(r"$\Delta\lambda$")
        plt.xlabel(f"Wavelength ({w.unit})")
        plt.ylabel(r"$\Delta\lambda$")
        plt.legend()

        plt.sca(ax[1])
        plt.plot(w, R_per_time, alpha=0.2)
        plt.plot(
            shared_w,
            shared_R / supersampling,
            color="black",
            label=f"(shared)/{supersampling}",
            marker=".",
        )
        plt.title(r"R = $\lambda/\Delta\lambda$")
        plt.xlabel(f"Wavelength ({w.unit})")
        plt.ylabel(r"R = $\lambda/\Delta\lambda$")
        plt.legend()
    return shared_w


def align_wavelengths(self, minimum_acceptable_ok=1, minimum_points_per_bin=0, **kw):
    """
    Use 2D wavelength information to align onto a single 1D wavelength array.

    Parameters
    ----------
    minimum_acceptable_ok : float
        The numbers in the `.ok` attribute express "how OK?" each
        data point is, ranging from 0 (not OK) to 1 (super OK).
        In most cases, `.ok` will be binary, but there may be times
        where it's intermediate (for example, if a bin was created
        from some data that were not OK and some that were).
        The `minimum_acceptable_ok` parameter allows you to specify what
        level of OK-ness for a point to go into the binning.
        Reasonable options may include:
            minimum_acceptable_ok = 1
                  Only data points that are perfectly OK
                  will go into the binning. All other points
                  will effectively be interpolated over. Flux
                  uncertainties *should* be inflated appropriately,
                  but it's very possible to create correlated
                  bins next to each other if many of your ingoing
                  data points are not perfectly OK.
            minimum_acceptable_ok = 1
                  All data points that aren't definitely not OK
                  will go into the binning. The OK-ness of points
                  will propagate onward for future binning.
            minimum_acceptable_ok < 0
                  All data points will be included in the bin.
                  The OK-ness will propagate onward.
    wscale : str
        What kind of a new wavelength axis should be created?
        Options include:
            'linear' = constant d[wavelength] between grid points
            'log' = constant d[wavelength]/[wavelength] between grid points

    supersampling : float
        By how many times should we increase or decrease the wavelength sampling?
        In general, values >1 will split each input wavelength grid point into
        multiple supersampled wavelength grid points, values close to 1 will
        produce approximately one output wavelength for each input wavelength,
        and values <1 will average multiple input wavelengths into a single output
        wavelength bin.

        Unless this is significantly less than 1, there's a good chance your output
        array may have strong correlations between one or more adjacent wavelengths.
        Be careful when trying to use the resulting uncertainties!

    """
    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("align_wavelengths", locals())

    if "wavelength_2d" not in self.fluxlike:
        cheerfully_suggest(
            f"""
        No 2D wavelength information was found, so
        it's assumed wavelengths don't need to be aligned.
        Wavelength alignment is being skipped!
        """
        )
        shifted = self._create_copy()
    else:
        # create a shared wavelength array
        shared_wavelengths = self._create_shared_wavelength_axis(**kw)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            # bin the rainbow onto that new grid, starting from 2D wavelengths
            shifted = self.bin_in_wavelength(
                wavelength=shared_wavelengths,
                minimum_acceptable_ok=minimum_acceptable_ok,
                starting_wavelengths="2D",
                minimum_points_per_bin=minimum_points_per_bin,
            )

    # append the history entry to the new Rainbow
    shifted._record_history_entry(h)

    # return the new Rainbow
    return shifted



from .withmodel import *


class SimulatedRainbow(RainbowWithModel):
    """
    `SimulatedRainbow` objects are created from scratch
    within `chromatic`, with options for various different
    wavelength grids, time grids, noise sources, and injected
    models. They can be useful for generating quick simulated
    dataset for testing analysis and visualization tools.

    This class definition inherits from `RainbowWithModel`,
    which itself inherits from `Rainbow`.
    """

    def __init__(
        self,
        tlim=[-2.5, 2.5] * u.hour,
        dt=2 * u.minute,
        time=None,
        wlim=[0.5, 5] * u.micron,
        R=100,
        dw=None,
        wavelength=None,
        star_flux=None,
        name=None,
        signal_to_noise=None,
    ):
        """
        Create a simulated rainbow object.

        Parameters
        ----------

        tlim : list or array of astropy.units.Quantity
            The pip install -e '.[develop]'[min, max] times for creating the time grid.
            These should have astropy units of time.
        dt : astropy.units.Quantity
            The d(time) bin size for creating a grid
            that is uniform in linear space.
        time : array of astropy.units.Quantity
            An array of times, if you just want to give
            it an entirely custom array.

        The time-setting order of precendence is:
            1) time
            2) tlim + dt

        wlim : list or array of astropy.units.Quantity
            The [min, max] wavelengths for creating the grid.
            These should have astropy units of wavelength.
        R : float
            The spectral resolution for creating a grid
            that is uniform in logarithmic space.
        dw : astropy.units.Quantity
            The d(wavelength) bin size for creating a grid
            that is uniform in linear space.
        wavelength : array of astropy.units.Quantity
            An array of wavelengths, if you just want to give
            it an entirely custom array.

        The wavelength-setting order of precendence is:
            1) wavelength
            2) wlim + dw
            3) wlim + R

        star_flux : numpy 1D array
            An array of fluxes corresponding to the supplied wavelengths.
            If left blank, the code assumes a normalized flux of
            flux(wavelength) = 1 for all wavelengths.
        """
        Rainbow.__init__(self)

        # (remove the history entry from creating the Rainbow)
        self._remove_last_history_entry()

        # create a history entry for this action (before other variables are defined)
        h = self._create_history_entry("SimulatedRainbow", locals())

        # set up the wavelength grid
        self._setup_fake_wavelength_grid(wlim=wlim, R=R, dw=dw, wavelength=wavelength)

        # set up the time grid
        self._setup_fake_time_grid(tlim=tlim, dt=dt, time=time)

        # save the basic inputs that aren't stored elsewhere
        self.metadata["name"] = name

        # If the flux of the star is not given,
        # assume a continuum-normlized flux where fx=1 at all wavelengths.
        if star_flux is None:
            model = np.ones(self.shape)

        # If the flux vs wavelength of the star is supplied,
        # include it in the model.
        else:
            # Check to make sure the flux and wavelengths
            # have the same shape.
            if len(star_flux) == len(self.wavelike["wavelength"]):
                model = np.transpose([star_flux] * self.shape[1])
            elif len(star_flux) == 1:
                model = star_flux * np.ones(self.shape)

        # Set uncertainty.
        self.fluxlike["flux"] = model * 1
        self.fluxlike["model"] = model * 1
        self.fluxlike["uncertainty"] = np.zeros(self.shape)

        # make sure everything is defined and sorted
        self._validate_core_dictionaries()

        if signal_to_noise is not None:
            message = f"""
            You tried to specify the noise level with
            `SimulatedRainbow(signal_to_noise={signal_to_noise})`,
            but that functionality is going away soon.
            Please replace it right now with
            `SimulatedRainbow().inject_noise(signal_to_noise={signal_to_noise})`
            so that your code will continue to work.
            You're getting away with it this time,
            but it won't work for much longer!
            """
            cheerfully_suggest(message)
            new = self.inject_noise()
            for k in ["flux", "uncertainty", "model"]:
                self.fluxlike[k] = new.fluxlike[k]

        # append the history entry to the new Rainbow
        self._record_history_entry(h)

    def _setup_fake_time_grid(
        self, tlim=[-2.5 * u.hour, 2.5 * u.hour], dt=1 * u.minute, time=None
    ):
        """
        Create a fake time grid.

        Parameters
        ----------

        tlim : list or array of astropy.units.Quantity
            The [min, max] times for creating the time grid.
            These should have astropy units of time.
        dt : astropy.units.Quantity
            The d(time) bin size for creating a grid
            that is uniform in linear space.
        time : array of astropy.units.Quantity
            An array of times, if you just want to give
            it an entirely custom array.

        The time-setting order of precendence is:
            1) time
            2) tlim + dt
        """
        # check we're trying to do exactly one thing
        if (tlim is None) and (time is None):
            raise RuntimeError("Please specify either `tlim` or `time`.")

        if time is None:
            t_unit = tlim[0].unit
            t_unit.to("s")
            time = np.arange(tlim[0] / t_unit, tlim[1] / t_unit, dt / t_unit) * t_unit
        else:
            t_unit = time.unit

        self.timelike["time"] = u.Quantity(time).to(u.day)
        # TODO, make this match up better with astropy time

        self._guess_tscale()

    def _setup_fake_wavelength_grid(
        self, wlim=[0.5 * u.micron, 5 * u.micron], R=100, dw=None, wavelength=None
    ):
        """
        Create a fake wavelength grid.

        Parameters
        ----------

        wlim : list or array of astropy.units.Quantity
            The [min, max] wavelengths for creating the grid.
            These should have astropy units of wavelength.
        R : float
            The spectral resolution for creating a grid
            that is uniform in logarithmic space.
        dw : astropy.units.Quantity
            The d(wavelength) bin size for creating a grid
            that is uniform in linear space.
        wavelength : array of astropy.units.Quantity
            An array of wavelengths, if you just want to give
            it an entirely custom array.

        The wavelength-setting order of precendence is:
            1) wavelength
            2) wlim + dw
            3) wlim + R
        """

        # check we're trying to do exactly one thing
        if (wlim is None) and (wavelength is None):
            raise RuntimeError("Please specify either `wlim` or `wavelength`.")

        # create a linear or logarithmic grid
        if wavelength is None:
            # check that we're
            if (R is None) and (dw is None):
                raise RuntimeError("Please specify either `R` or `dw`.")

            w_unit = wlim[0].unit
            if dw is None:
                self.metadata["R"] = R
                # self.metadata["wscale"] = "log"

                logw_min = np.log(wlim[0] / w_unit)
                logw_max = np.log(wlim[1] / w_unit)
                logw = np.arange(logw_min, logw_max, 1 / R)
                wavelength = np.exp(logw) * w_unit

            elif dw is not None:
                self.metadata["dw"] = dw
                # self.metadata["wscale"] = "linear"
                wavelength = (
                    np.arange(wlim[0] / w_unit, wlim[1] / w_unit, self.dw / w_unit)
                    * w_unit
                )

        # or just make sure the wavelength grid has units
        elif wavelength is not None:
            w_unit = wavelength.unit

        # make sure the wavelength array has units
        self.wavelike["wavelength"] = u.Quantity(wavelength).to(u.micron)
        self._guess_wscale()



from ...imports import *
from ..multi import *

__all__ = ["compare"]


def compare(self, rainbows):
    """
    Compare this Rainbow to others.

    Parameters
    ----------
    rainbows : list
        A list containing one or more other Rainbow objects.
        If you only want to compare with one other Rainbow,
        supply it in a 1-element list like `.compare([other])`
    """
    try:
        rainbows.remove(self)
    except (ValueError, IndexError):
        pass
    return compare_rainbows([self] + rainbows)



from ..rainbows import *
from .setup_tests import *


def test_multi():
    a = (
        SimulatedRainbow(R=3, dt=20 * u.minute)
        .inject_transit()
        .inject_noise(signal_to_noise=400)
    )
    b = (
        SimulatedRainbow(R=3, dt=20 * u.minute)
        .inject_transit()
        .inject_noise(signal_to_noise=800)
    )

    m = MultiRainbow([a, b])
    m.plot(spacing=0.02)
    plt.savefig(os.path.join(test_directory, "multi-plot-demonstration.png"))

    m.imshow()
    plt.savefig(os.path.join(test_directory, "multi-imshow-demonstration.png"))

    m.animate_lightcurves(
        filename=os.path.join(
            test_directory, "multi-animate-lightcurves-demonstration.gif"
        )
    )
    m.animate_spectra(
        filename=os.path.join(test_directory, "multi-animate-spectra-demonstration.gif")
    )

    m.normalize()
    # m.align_wavelengths().wavelength
    # m[:, :]
    plt.close("all")


def test_compare_wrappers():
    rainbows = [SimulatedRainbow(R=10 ** np.random.uniform(0.1, 2)) for _ in range(3)]

    a = compare_rainbows(rainbows)
    b = rainbows[0].compare(rainbows)

    assert a.names == b.names
    assert a.rainbows == b.rainbows



from .x1dints import *
from .x1dints_kludge import *

from .eureka_txt import *
from .eureka_specdata import *
from .eureka_lcdata import *
from .eureka_channels import *
from .rainbow_npy import *
from .rainbow_FITS import *
from .text import *

from .xarray_stellar_spectra import *
from .xarray_raw_light_curves import *
from .xarray_fitted_light_curves import *

# some particular instruments
from .nres import *
from .atoca import *

# some individual-ish folks
from .espinoza import *
from .dossantos import *
from .feinstein import *
from .schlawin import *
from .coulombe import *
from .kirk import *
from .radica import *


# construct a dictionary of available readers
available_readers = {k: globals()[k] for k in globals() if k[0:5] == "from_"}


def guess_reader(filepath, format=None):
    """
    A wrapper to guess the appropriate reader from the filename
    (and possibily an explicitly-set file format string).

    Parameters
    ----------
    filepath : str
        The path to the file, or a glob-friendly string
        pointing to a group of files that should all be
        loaded together (for example, if an exposure was
        split into multiple segments).
    format : str, None
        The file format to use.
    """

    from ...imports import expand_filenames
    from fnmatch import fnmatch
    import glob

    # get all the possible filenames (= expand wildcard)
    filenames = expand_filenames(filepath)
    f = filenames[0]

    # if format='abcdefgh', return the `from_abcdefgh` function
    if format is not None:
        return available_readers[f"from_{format}"]
    # does it look like a .rainbow.npy chromatic file?
    elif fnmatch(f, "*.rainbow.npy"):
        return from_rainbow_npy
    elif fnmatch(f.lower(), "*.rainbow.fits") or fnmatch(f.lower(), "*.rainbow.fit"):
        return from_rainbow_FITS
    # does it look like an ERS-xarray format?
    elif fnmatch(f, "*stellar-spec*.xc"):
        return from_xarray_stellar_spectra
    # does it look like an ERS-xarray format?
    elif fnmatch(f, "*raw-light-curve*.xc"):
        return from_xarray_raw_light_curves
    # does it look like an ERS-xarray format?
    elif fnmatch(f, "*fitted-light-curve*.xc"):
        return from_xarray_fitted_light_curves
    # does it look like a .rainbow.npy chromatic file?
    elif fnmatch(f, "*order*.npy"):
        return from_espinoza
    # does it look like a STScI x1dints.fits file?
    elif fnmatch(f, "*x1dints.fits"):
        return from_x1dints
    # does it look like a STScI x1dints.fits file?
    elif fnmatch(f, "*extract_1d.fits"):
        return from_x1dints_kludge
    # does it look like an Eureka! S3 text file?
    elif fnmatch(f, "*S3_*_Save.dat") or fnmatch(f, "*S3_*_Save.txt"):
        return from_eureka_S3_txt
    # does it look like an Eureka! S3 SpecData hdf5 file?
    elif fnmatch(f, "*S3_*_SpecData.h5"):
        return from_eureka_S3
    # does it look like an Eureka! S4 SpecData hdf5 file?
    elif fnmatch(f, "*S4_*_SpecData.h5"):
        return from_eureka_SpecData
    # does it look like an Eureka! S4 LCData hdf5 file?
    elif fnmatch(f, "*S4_*_LCData.h5"):
        return from_eureka_S4
    elif fnmatch(f, "*S5_*_Table_Save_*.txt"):
        return from_eureka_S5
    elif fnmatch(f, "*extract1dstep.fits"):
        return from_atoca
    elif fnmatch(f, "*wb_lcs*"):
        return from_kirk_fitted_light_curves
    elif fnmatch(f, "*_flux_resampled*.pickle"):
        return from_kirk_stellar_spectra
    elif fnmatch(f, "*.txt") or fnmatch(f, "*.csv"):
        return from_text
    elif fnmatch(f, "*e92-1d.fits") or fnmatch(f, "*e92-1d.fits.fz"):
        return from_nres
    elif fnmatch(f, "*spec_*.fits"):
        return from_schlawin
    else:
        raise ValueError(
            f"""
        We're having trouble guessing the input format from the filename
        {f}
        Please try specifying a `format=` keyword to your call.
        """
        )



from ..rainbows import *
from .setup_tests import *


def test_normalize(plot=False):
    N = 37
    w = np.logspace(0, 1, N) * u.micron
    f = np.cos(2 * np.pi * w.value / 3) + 1

    snr = 100
    a = SimulatedRainbow(wavelength=w).inject_noise(signal_to_noise=snr)
    b = SimulatedRainbow(wavelength=w, star_flux=f).inject_noise(signal_to_noise=snr)
    c = (
        SimulatedRainbow(wavelength=w, star_flux=f)
        .inject_transit()
        .inject_noise(signal_to_noise=snr)
    )

    for x in [a, b, c]:
        nw = x.normalize(axis="w")
        nt = x.normalize(axis="t")
        nwt = x.normalize(axis="w").normalize(axis="t")
        ntw = x.normalize(axis="t").normalize(axis="w")

        for r in [nw, nt, nwt, ntw]:
            r.fluxlike["relative-uncertainty"] = r.uncertainty / r.flux
            assert np.all(np.isclose(r.uncertainty / r.flux, 1 / snr, rtol=0.1))
            if plot:
                r.imshow_quantities(maxcol=4)
    plt.close("all")


def test_is_probably_normalized():
    f = [2] * u.W / u.m**2
    kw = dict(star_flux=f, R=10, dt=10 * u.minute)
    assert SimulatedRainbow(**kw)._is_probably_normalized() == False
    assert SimulatedRainbow(**kw).normalize()._is_probably_normalized() == True
    assert SimulatedRainbow(**kw).inject_noise()._is_probably_normalized() == False
    assert (
        SimulatedRainbow(**kw).inject_noise().normalize()._is_probably_normalized()
        == True
    )



from ...imports import *


def _is_probably_rainbow(x):
    """
    A quick check if an object is probably a Rainbow
    (without having to do a recursive import to compare
    directly to the class definition).
    """
    return "Rainbow" in x.__class__.__name__


def _do_rainbows_match(a, b):
    """
    A check that two Rainbows have matching wavelengths and times.
    If either of the others aren't Rainbows, an AssertionError
    will be raised.

    Parameters
    ----------
    a : Rainbow
        One Rainbow other.
    b : Rainbow
        Another Rainbow other.

    Returns
    -------
    match : bool
        Do their wavelengths and times match?
    """

    return np.array_equal(a.wavelength, b.wavelength) and np.array_equal(a.time, b.time)


def _raise_ambiguous_shape_error(self, x):
    """
    Raise an error if the shape of a Rainbow is ambiguous.
    """
    if x.nwave == x.ntime:
        raise RuntimeError(
            f"{self} has same number of wavelengths and times; we can't tell which is which."
        )


def _broadcast_to_fluxlike(self, x):
    """
    Convert a scalar, wavelike array, timelike array, or
    fluxlike array into something that can be broadcast into
    a fluxlike shape for math.

    Parameters
    ----------
    x : int, float, bool, array, u.Quantity, ...
        The input.

    Returns
    -------
    x[:,:] : array
        The input in a format that can happily do
        math with a 2D fluxlike array.
    """
    if np.shape(x) == (self.nwave,):
        self._raise_ambiguous_shape_error(self)
        return x[:, np.newaxis]
    elif np.shape(x) == (self.ntime,):
        self._raise_ambiguous_shape_error(self)
        return x[np.newaxis, :]
    elif np.shape(x) in [(), (1,), self.shape]:
        return x
    else:
        raise ValueError(
            f"The shapes {np.shape(x)} and {self.shape} cannot be cast together."
        )


def _apply_operation(self, other, operation, dzdx="1", dzdy="1"):
    """
    Apply a mathematical operation between `self` (x) and `other` (y)
    that returns a new Rainbow where the `flux` (and maybe `model`)
    are the result of the operation (+, -, *, /). Uncertainties
    will be propagated using the string expressions for the required
    derivatives.

    Parameters
    ----------
    self : Rainbow
        The current Rainbow object.
    other : Rainbow, array, float, int
        There are multiple options for what this could be:
            1) scalar float or int
            2) 1D array with same length as wavelength axis
            3) 1D array with same length as time axis
            4) 2D array with same shape as rainbow flux
            5) Rainbow other with same dimensions as self.
    operation : function
        The function that will be applied to combine `self` and `other`.
        Likely options are `np.add`, `np.subtract`, `np.multiply`, and
        `np.true_divide`.
    dzdx : string
        For propagating uncertainties, if z(x,y) is the result,
        what is `dz/dx`? This must be supplied as a string that
        uses the variables `x` (= `self`) and `y` (= `other`).
    dzdy : string
        For propagating uncertainties, if z(x,y) is the result,
        what is `dz/dy`? This must be supplied as a string that
        uses the variables `x` (= `self`) and `y` (= `other`).
    """
    # create new Rainbow() to store results in.
    result = self._create_copy()

    # other object is not Rainbow
    if _is_probably_rainbow(other):
        if _do_rainbows_match(self, other):
            for k in self._keys_that_respond_to_math:
                result.fluxlike[k] = operation(self.fluxlike[k], other.fluxlike[k])
        else:
            raise ValueError(
                f"The two Rainbow objects {self} and {other} don't share wavelength/time axes."
            )
        sigma_y = other.uncertainty
        y = other.get("model")
        if y is None:
            y = other.flux

    # other object not Rainbow
    else:
        y = self._broadcast_to_fluxlike(other)
        sigma_y = 0
        for k in self._keys_that_respond_to_math:
            result.fluxlike[k] = operation(self.fluxlike[k], y)

    sigma_x = self.uncertainty
    x = self.get("model")
    if x is None:
        x = self.flux

    # If z = operation(x,y), then to propagate errors we need
    # to use the derivatives of z with respect to x (dz/dx) and y (dz/dy):
    # print(f"mean(x) = {np.mean(x)}")
    # print(f"mean(sigma_x) = {np.mean(sigma_x)}")
    # print(f"dzdx = {dzdx}")
    # print(f"mean(y) = {np.mean(y)}")
    # print(f"mean(sigma_y) = {np.mean(sigma_y)}")
    # print(f"dzdy = {dzdy}")

    variance = sigma_x**2 * eval(dzdx) ** 2 + sigma_y**2 * eval(dzdy) ** 2
    result.fluxlike["uncertainty"] = np.sqrt(variance)

    return result


def __add__(self, other):
    """
    Add the flux of a rainbow and an input array (or another rainbow)
    and output in a new rainbow other.
    Currently only supports flux addition.

    Parameters
    ----------
    other : Array or float.
        Multiple options:
        1) float
        2) 1D array with same length as wavelength axis
        3) 1D array with same length as time axis
        4) 2D array with same shape as rainbow flux
        5) Rainbow other with same dimensions as self.

    Returns
    ----------
    rainbow : Rainbow() with the same parameters as self but with added
    flux.
    """

    # create the history entry
    h = self._create_history_entry("+", locals())

    # calculate a new Rainbow using the operation and error propagation
    result = self._apply_operation(other, operation=np.add, dzdx="1", dzdy="1")

    # append the history entry to the new Rainbow
    result._record_history_entry(h)

    return result


def __sub__(self, other):
    """
    Subtract the flux of a rainbow from an input array (or another rainbow)
    and output in a new rainbow other.
    Currently only supports flux subtraction.

    Parameters
    ----------
    other : Array or float.
        Multiple options:
        1) float
        2) 1D array with same length as wavelength axis
        3) 1D array with same length as time axis
        4) 2D array with same shape as rainbow flux
        5) Rainbow other with same dimensions as self.

    Returns
    ----------
    rainbow : Rainbow() with the same parameters as self but with subtracted
    flux.
    """
    # create the history entry
    h = self._create_history_entry("-", locals())

    # calculate a new Rainbow using the operation and error propagation
    result = self._apply_operation(other, operation=np.subtract, dzdx="1", dzdy="1")

    # append the history entry to the new Rainbow
    result._record_history_entry(h)

    return result


def __mul__(self, other):
    """
    Multiply the flux of a rainbow and an input array (or another rainbow)
    and output in a new rainbow other.
    Currently only supports flux multiplication.

    Parameters
    ----------
    other : Array or float.
        Multiple options:
        1) float
        2) 1D array with same length as wavelength axis
        3) 1D array with same length as time axis
        4) 2D array with same shape as rainbow flux
        5) Rainbow other with same dimensions as self.

    Returns
    ----------
    rainbow : Rainbow() with the same parameters as self but with multiplied
    flux.
    """

    # create the history entry
    h = self._create_history_entry("*", locals())

    # calculate a new Rainbow using the operation and error propagation
    result = self._apply_operation(other, operation=np.multiply, dzdx="y", dzdy="x")

    # append the history entry to the new Rainbow
    result._record_history_entry(h)

    return result


def __truediv__(self, other):
    """
    Divide the flux of a rainbow and an input array (or another rainbow)
    and output in a new rainbow other.
    Currently only supports flux division.

    Parameters
    ----------
    other : Array or float.
        Multiple options:
        1) float
        2) 1D array with same length as wavelength axis
        3) 1D array with same length as time axis
        4) 2D array with same shape as rainbow flux
        5) Rainbow other with same dimensions as self.

    Returns
    ----------
    rainbow : Rainbow() with the same parameters as self but with divided
    flux.
    """
    # create the history entry
    h = self._create_history_entry("/", locals())

    # calculate a new Rainbow using the operation and error propagation
    result = self._apply_operation(
        other, operation=np.true_divide, dzdx="1/y", dzdy="-x/y**2"
    )

    # append the history entry to the new Rainbow
    result._record_history_entry(h)

    return result


def __eq__(self, other):
    """
    Test whether (self) == (other).
    This compares the wavelike, timelike, and fluxlike arrays
    for exact matches. It skips entirely over the metadata.
    """
    # start by assuming the Rainbows are identical
    same = True

    for a, b in zip([self, other], [other, self]):

        # loop through the core dictionaries
        for d in a._core_dictionaries:
            if d == "metadata":
                continue
            # pull out each core dictionary from both
            d1, d2 = vars(self)[d], vars(b)[d]
            same *= set(d1.keys()) == set(d2.keys())

            # loop through elements of each dictionary
            for k in d1:

                # ignore different histories (e.g. new vs loaded)
                if k != "history":

                    # test that all elements match for both
                    if d == "fluxlike":
                        same *= np.all(
                            np.isclose(
                                a.get(k)[a.ok.astype(bool)], b.get(k)[a.ok.astype(bool)]
                            )
                        )
                    else:
                        same *= np.all(np.isclose(a.get(k), b.get(k)))

    return bool(same)


def diff(self, other):
    """
    Test whether (self) == (other).
    This compares the wavelike, timelike, and fluxlike arrays
    for exact matches. It skips entirely over the metadata.
    """
    # start by assuming the Rainbows are identical
    same = True

    for a, b in zip([self, other], [other, self]):

        # loop through the core dictionaries
        for d in a._core_dictionaries:
            if d == "metadata":
                continue
            # pull out each core dictionary from both
            d1, d2 = vars(self)[d], vars(b)[d]
            if set(d1.keys()) != set(d2.keys()):
                differences = list(set(d1.keys()) - set(d2.keys()))
                print(f"{a}.{d} has {differences} and {b} does not")

            # loop through elements of each dictionary
            for k in d1:
                # ignore different histories (e.g. new vs loaded)
                if k != "history":
                    continue
                # test that all elements match for both
                if np.all(a.get(k) == b.get(k)) == False:
                    print(f"{a}.{d}[{k}] != {b}.{d}[{k}]")


0



from ...imports import *


def attach_model(self, model, **kw):
    """
    Attach a fluxlike model to this Rainbow,
    storing it as `self.fluxlike['model']`

    After running this once, it's OK (and faster) to simply
    update the `.model` attribute of the result.

    Parameters
    ----------
    model : np.array, u.Quantity
        An array of model values, with the same shape as 'flux'
    kw : dict
        All other keywords will be interpreted as items
        that can be added to a Rainbow. You might use this
        to attach intermediate model steps or quantities
        (for example, 'planet_model' or 'systematics_model').
    """

    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("attach_model", locals())

    # make sure the shape is reasonable
    assert np.shape(model) == np.shape(self.flux)

    # add the model to the fluxlike array
    inputs = self._create_copy()._get_core_dictionaries()
    inputs["fluxlike"]["model"] = model

    # import here (rather than globally) to avoid recursion?
    from ..withmodel import RainbowWithModel

    # create new object
    new = RainbowWithModel(**inputs)

    # add other inputs to the model
    for k, v in kw.items():
        new.__setattr__(k, v)

    # append the history entry to the new Rainbow
    new._record_history_entry(h)

    # return the RainboWithModel
    return new



"""
Define a writer for chromatic .rainbow.FITS files.
"""

# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["to_rainbow_FITS"]


def to_rainbow_FITS(self, filepath, overwrite=True):
    """
    Write a Rainbow to a FITS file.

    Parameters
    ----------

    self : Rainbow
        The object to be saved.

    filepath : str
        The path to the file to write.
    """

    # create a header for the metadata
    header = fits.Header()
    header["comment"] = "Data are stored in three main extensions"
    header["comment"] = " [1]=['FLUXLIKE'] contains (nwave, ntime)-shaped arrays"
    header["comment"] = " [2]=['WAVELIKE'] contains (nwave)-shaped arrays"
    header["comment"] = " [3]=['TIMELIKE'] contains (ntime)-shaped arrays"
    header["comment"] = "The primary extension header contains some metadata."

    for k in self.metadata:
        try:
            header[k] = self.metadata[k]
        except ValueError:
            cheerfully_suggest(f"metadata item '{k}' cannot be saved to FITS header")

    primary_hdu = fits.PrimaryHDU(header=header)

    # create extensions for the three other core dictionaries
    flux_hdu = fits.BinTableHDU(Table(self.fluxlike), name="fluxlike")
    wave_hdu = fits.BinTableHDU(Table(self.wavelike), name="wavelike")
    time_hdu = fits.BinTableHDU(Table(self.timelike), name="timelike")

    hdu_list = fits.HDUList([primary_hdu, flux_hdu, wave_hdu, time_hdu])
    hdu_list.writeto(filepath, overwrite=overwrite)



from ..rainbows import *
from .setup_tests import *


def test_astropy_times():
    s = SimulatedRainbow(time=(Time.now().jd + np.linspace(0, 1)) * u.day)
    format = "jd"
    scale = "tdb"
    original_times = s.time
    astropy_times = s.get_times_as_astropy(
        format=format, scale=scale, is_barycentric=True
    )
    s.set_times_from_astropy(astropy_times, is_barycentric=True)

    assert np.all(s.time == original_times)
    assert np.all(astropy_times == s.get_times_as_astropy())



from ...imports import *
import batman

__all__ = [
    "inject_systematics",
    "_create_fake_wavelike_quantity",
    "_create_fake_timelike_quantity",
    "_create_fake_fluxlike_quantity",
]

"""
The idea for simple GP-generated systematics came from the fabulous
tutorial Neale Gibson led for the ers-transit team in summer 2021

The recording and a follow-along notebook are available at
https://ers-transit.github.io/pre-launch-hackathon.html
"""


def create_covariance_matrix(x, correlated=1, length=3, uncorrelated=0.5):
    """
    Generate a covariance matrix from a squared-exponential kernel
    with an additional diagonal component. The kernel is given by
    the following expression for x[m] and x[n]:

    distance = x[m] - x[n]
    correlated**2 * exp(-0.5*distance**2/length**2) + uncorrelated**2 * delta(m,n)

    It will be normalized to correlated**2 + uncorrelated**2 = 1

    Parameters
    ----------
    x : np.array or u.Quantity
        The arry of values for generating the matrix.
    correlated : float
        The amplitude of the correlated noise component.
    length : float or u.Quantity
        The length scale of the correlation (in same units as x)
    uncorrelated : float
        The amplitude of the uncorrelated noise component.
    """

    # what's the distance between each point and each other (an square matrix)
    distance_squared = np.subtract.outer(x, x) ** 2

    # calculate normalized coefficients
    N = correlated**2 + uncorrelated**2
    A, B = correlated / N, uncorrelated / N

    # generate the covariance matrix with squared exponential kernel
    covariance_matrix = A**2 * np.exp(-0.5 * distance_squared / length**2)

    # add extra variance along the diagonal
    i = np.arange(len(x))
    covariance_matrix[i, i] += B**2

    return covariance_matrix


def _create_fake_timelike_quantity(
    self, mean=np.zeros_like, correlated=1, dt=0.5 * u.hour, uncorrelated=0.25
):
    """
    Create a fake timelike quantity.

    Parameters
    ----------
    mean : function
        A function that takes in an array of times and returns
        a mean value for each time. The fake timeseries will
        wobble around that mean, according to the covariance
        matrix.
    correlated : float
        The amplitude of the correlated noise component.
    dt : u.Quantity
        The time scale of the correlation (in units of time).
    uncorrelated : float
        The amplitude of the uncorrelated noise component.
        (correlated and uncorrelated will be normalized to 1)
    Returns
    -------
    timelike : np.array
        A timelike array with the requested noise properties.
    """
    x = self.time
    covariance_matrix = create_covariance_matrix(
        x, correlated=correlated, length=dt, uncorrelated=uncorrelated
    )
    return np.random.multivariate_normal(mean(x), covariance_matrix)


def _create_fake_wavelike_quantity(
    self, mean=np.zeros_like, correlated=1, R=5, uncorrelated=0.25
):
    """
    Create a fake timelike quantity.

    Parameters
    ----------
    mean : function
        A function that takes in an array of times and returns
        a mean value for each time. The fake timeseries will
        wobble around that mean, according to the covariance
        matrix.
    correlated : float
        The amplitude of the correlated noise component.
    dt : u.Quantity
        The time scale of the correlation (in units of time).
    uncorrelated : float
        The amplitude of the uncorrelated noise component.
        (correlated and uncorrelated will be normalized to 1)

    Returns
    -------
    wavelike : np.array
        A wavelike array with the requested noise properties.
    """
    x = np.log(self.wavelength.value)
    covariance_matrix = create_covariance_matrix(
        x, correlated=correlated, length=1 / R, uncorrelated=uncorrelated
    )
    return np.random.multivariate_normal(mean(x), covariance_matrix)


def _create_fake_fluxlike_quantity(self, timelike_kw={}, wavelike_kw={}):
    """
    Create a fake fluxlike quantity.

    Parameters
    ----------
    timelike_kw : dict
        Dictionary of keywords to pass to `_create_fake_timelike_quantity()`

    Returns
    -------
    fluxlike : np.array
        A fluxlike array (with )
    """

    # create a 1D time array
    tkw = dict(correlated=1, uncorrelated=0.5, dt=1 * u.hour)
    tkw.update(**timelike_kw)
    t = self._create_fake_timelike_quantity(**tkw)

    # create a 1D wavelength array
    wkw = dict(correlated=1, uncorrelated=0, R=5)
    wkw.update(**wavelike_kw)
    w = self._create_fake_wavelike_quantity(**wkw)

    # multiply them together and return
    f = w[:, np.newaxis] * t[np.newaxis, :]
    return f


def inject_systematics(
    self,
    amplitude=0.003,
    wavelike=[],
    timelike=["x", "y", "time"],
    fluxlike=["background"],
):

    """
    Inject some (very cartoony) instrumental systematics.

    Here's the basic procedure:

    1) Generate some fake variables that vary either just with
    wavelength, just with time, or with both time and wavelength.
    Store these variables for later use. For example, these might
    represent an average `x` and `y` centroid of the trace on the
    detector (one for each time), or the background flux associated
    with each wavelength (one for each time and for each wavelength).

    2) Generate a flux model as some function of those variables.
    In reality, we probably don't know the actual relationship
    between these inputs and the flux, but we can imagine one!

    3) Inject the model flux into the `flux` of this Rainbow,
    and store the combined model in `systematics-model` and
    each individual component in `systematic-model-{...}`.

    Parameters
    ----------
    amplitude : float
        The (standard deviation-ish) amplitude of the systematics
        in units normalized to 1. For example, an amplitude of 0.003
        will produce systematic trends that tend to range (at 1 sigma)
        from 0.997 to 1.003.
    wavelike : list of strings
        A list of wave-like cotrending quantities to serve as ingredients
        to a linear combination systematics model. Existing quantities
        will be pulled from the appropriate core dictionary; fake
        data will be created for quantities that don't already exist,
        from a cartoony Gaussian process model.
    timelike : list of strings
        A list of time-like cotrending quantities to serve as ingredients
        to a linear combination systematics model. Existing quantities
        will be pulled from the appropriate core dictionary; fake
        data will be created for quantities that don't already exist,
        from a cartoony Gaussian process model.
    fluxlike : list of strings
        A list of flux-like cotrending quantities to serve as ingredients
        to a linear combination systematics model. Existing quantities
        will be pulled from the appropriate core dictionary; fake
        data will be created for quantities that don't already exist,
        from a cartoony Gaussian process model.

    Returns
    -------
    rainbow : Rainbow
        A new Rainbow object with the systematics injected.
    """

    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("inject_systematics", locals())

    # create a copy of the existing Rainbow
    new = self._create_copy()
    new.fluxlike["systematics_model"] = np.ones(self.shape)

    def standardize(q):
        """
        A quick helper to normalize all inputs to zero mean
        and unit standard deviation. It
        """
        offset = np.nanmean(q)
        sigma = np.nanstd(q)
        return u.Quantity((q - offset) / sigma).value, offset, sigma

    components = {}
    for k in wavelike:
        if k in self.wavelike:
            x, offset, sigma = standardize(self.wavelike[k])
        else:
            x = new._create_fake_wavelike_quantity()
            offset, sigma = 0, 1
            new.wavelike[k] = x
        c = np.random.normal(0, amplitude)
        df = c * x[:, np.newaxis] * np.ones(self.shape)
        new.fluxlike[f"systematics_model_from_{k}"] = df
        new.fluxlike["systematics_model"] += df
        components.update(
            **{
                f"linear_{k}": f"c_{k}*({k} - offset_{k})/sigma_{k}",
                f"c_{k}": c,
                f"offset_{k}": offset,
                f"sigma_{k}": sigma,
            }
        )

    for k in timelike:
        if k in self.timelike:
            x, offset, sigma = standardize(self.timelike[k])
        else:
            x = new._create_fake_timelike_quantity()
            offset, sigma = 0, 1
            new.timelike[k] = x
        c = np.random.normal(0, amplitude)
        df = c * x[np.newaxis, :] * np.ones(self.shape)
        new.fluxlike[f"systematics_model_from_{k}"] = df
        new.fluxlike["systematics_model"] += df
        components.update(
            **{
                f"linear_{k}": f"c_{k}*({k} - offset_{k})/sigma_{k}",
                f"c_{k}": c,
                f"offset_{k}": offset,
                f"sigma_{k}": sigma,
            }
        )

    for k in fluxlike:
        if k in self.fluxlike:
            x, offset, sigma = standardize(self.fluxlike[k])
        else:
            x = new._create_fake_fluxlike_quantity()
            offset, sigma = 0, 1
            new.fluxlike[k] = x
        c = np.random.normal(0, amplitude)
        df = c * x * np.ones(self.shape)
        new.fluxlike[f"systematics_model_from_{k}"] = df
        new.fluxlike["systematics_model"] += df
        components.update(
            **{
                f"linear_{k}": f"c_{k}*({k} - offset_{k})/sigma_{k}",
                f"c_{k}": c,
                f"offset_{k}": offset,
                f"sigma_{k}": sigma,
            }
        )

    new.metadata["systematics_components"] = components
    new.metadata["systematics_equation"] = "f = 1\n  + " + "\n  + ".join(
        [v for k, v in components.items() if k[:7] == "linear_"]
    )

    # modify both the model and flux arrays
    new.flux *= new.systematics_model
    new.model = new.fluxlike.get("model", 1) * new.systematics_model

    # append the history entry to the new Rainbow
    new._record_history_entry(h)

    # return the new object
    return new



from ....imports import *

__all__ = ["plot_quantities"]


def plot_quantities(
    self,
    quantities=None,
    xaxis="time",
    maxcol=1,
    panel_size=(6, 2),
    what_is_x="index",
    filename=None,
    **kw,
):
    """
    Plot {xaxis}-like quantities as a function of {xaxis} index
    or any other {xaxis} quantity (such as "time" or "wavelength").

    Parameters
    ----------
    quantities : list like
        The {xaxis}-like quantity to plot.
    xaxis : string
        Whether the quantities are alike to 'time' or 'wave'. Default is 'time'. (Optional)
    maxcol : int
        The maximum number of columns to show (Optional).
    panel_size : tuple
        The size in inches dedicated to each panel, default is (6,2). (Optional)
    what_is_x : string
        The quantity to plot on the what_is_x, default is index. (Optional)

    """
    # decide which dictionary to plot
    if xaxis not in ["time", "wave", "wavelength"]:
        raise Exception("Unknown xaxis. Choose from [time, wave]")
    elif xaxis == "time":
        like_dict = self.timelike
    else:
        like_dict = self.wavelike

    # decide which quantities to plot
    if quantities is None:
        allkeys = like_dict.keys()
    else:
        allkeys = quantities[:]

    # set up the geometry of the grid
    if len(allkeys) > maxcol:
        rows = int(np.ceil(len(allkeys) / maxcol))
        cols = maxcol
    else:
        rows = 1
        cols = np.min([len(allkeys), maxcol])

    # create the figure and grid of axes
    fig, axes = plt.subplots(
        rows,
        cols,
        figsize=(cols * panel_size[0], rows * panel_size[1]),
        sharex=True,
        constrained_layout=True,
    )
    # make the axes easier to index
    if len(allkeys) > 1:
        ax = axes.flatten()
    else:
        ax = [axes]

    # set what_is_x variable
    if what_is_x.lower() == "index":
        x_values = np.arange(0, len(like_dict[list(like_dict.keys())[0]]))
        xlab = f"{xaxis} Index"
    else:
        if what_is_x in like_dict.keys():
            x_values = like_dict[what_is_x]
            xlab = what_is_x
        else:
            raise Exception("Desired `what_is_x` quantity is not in given dictionary")

    # display each quantity
    for k, key in enumerate(allkeys):
        # make the plot (or an empty box)
        if key in like_dict.keys():
            ax[k].plot(
                x_values, like_dict[key], color=plt.cm.viridis(k / len(allkeys)), **kw
            )
            ax[k].set_xlabel(xlab)
            ax[k].set_ylabel(
                f"{key} ({u.Quantity(like_dict[key]).unit.to_string('latex_inline')})"
            )
        else:
            ax[k].text(
                0.5,
                0.5,
                f"No {key}",
                transform=ax[k].transAxes,
                ha="center",
                va="center",
            )

        # add a title for each box
        ax[k].set_title(key)

        # hide xlabel except on the bottom row
        if k < (len(allkeys) - cols):
            ax[k].set_xlabel("")
        else:
            ax[k].tick_params(labelbottom=True)

    # hide any additional axes
    if k + 1 <= len(ax):
        for axi in ax[k + 1 :]:
            axi.axis("Off")
    if filename is not None:
        self.savefig(filename)



# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["from_schlawin"]


def from_schlawin(rainbow, filepath):
    """
    Populate a Rainbow from a file in Everett Schlawin's tshirt format.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.
    filepath : str
        The path to the file to load.
    """

    # read in your file, however you like
    hdu_list = fits.open(filepath)

    # populate a 1D array of wavelengths (with astropy units of length)
    rainbow.wavelike["indices"] = hdu_list["disp indices"].data * 1
    rainbow.wavelike["wavelength"] = hdu_list["wavelength"].data * u.micron * 1

    # populate a 1D array of times (with astropy units of time)
    rainbow.timelike["time"] = hdu_list["time"].data * u.day * 1

    # populate a 2D (row = wavelength, col = array of fluxes
    for k in [
        "optimal spec",
        "opt spec err",
        "sum spec",
        "sum spec err",
        "background spec",
        "refpix",
    ]:
        rainbow.fluxlike[k] = hdu_list[k].data.T.squeeze() * 1
    rainbow.fluxlike["flux"] = rainbow.fluxlike["optimal spec"] * 1
    rainbow.fluxlike["uncertainty"] = rainbow.fluxlike["opt spec err"] * 1

    # kludgily pull all extension headers into the metadata
    for k in hdu_list:
        rainbow.metadata[f"{k}-header"] = hdu_list[k].header



"""
This module serves as a template for creating a new Rainbow
reader. If you want to add the ability to read chromatic
light curves from a new kind of file format, a good process
would be to do something like the following

    1. Copy this `template.py` file into a new file in the
    `readers/` directory, ideally with a name that's easy
    to recognize, such as `readers/dossantos.py` (assuming
    `dossantos` is the name of your format)

    2. Start by finding and replacing `dossantos` in this
    template with the name of your format.

    3. Edit the `from_dossantos` function so that it will
    load a chromatic light curve file in your format and,
    for some Rainbow object `rainbow`, populate at least:

        + rainbow.timelike['time']
        + rainbow.wavelike['wavelength']
        + rainbow.fluxlike['flux']

    You'll need to replace the cartoon functions on each
    line with the actual code needed to load your file.

    (This template assumes that only one file needs to be
    loaded. If you need to load multiple segments, or each
    time point is stored in its own file or something, then
    check out `stsci.py` for an example of loading and
    stitching together multiple input files. You'll probably
    want to change `filepath` to accept a glob-friendly string
    like `my-neato-formatted-files-*.npy` or some such.)

    4. Edit the `readers/__init__.py` file to import your
    `from_dossantos` function to be accessible when people
    are trying to create new Rainbows. Add an `elif` statement
    to the `guess_reader` function that will help guess which
    reader to use from some aspect(s) of the filename.

    (This `guess_reader` function also accepts a `format=`
    keyword that allows the user to explicitly specify that
    the dossantos reader should be used.)

    5. Submit a pull request to the github repository for
    this package, so that other folks can use your handy
    new reader too!
"""

# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["from_dossantos"]


def from_dossantos(rainbow, filepath):
    """
    Populate a Rainbow from a file in the dossantos format.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated. This function is meant
        to be called during the initialization of a Rainbow
        object. You can/should assume that the `rainbow` object
        being passed will already contain the four core
        dictionaries `timelike`, `wavelike`, `fluxlike`, and
        `metadata`. This function should, at a minimum, add
        the following items
            + `rainbow.timelike['time']`
            + `rainbow.wavelike['wavelength']`
            + `rainbow.fluxlike['flux']`
        and optionally, additional entries like
            + `rainbow.metadata['some-useful-parameter']`
            + `rainbow.fluxlike['uncertainty']`
            + `rainbow.fluxlike['ok']`

    filepath : str
        The path to the file to load.
    """

    import pickle

    with open(filepath, "rb") as pkl:
        spectra = pickle.load(pkl)

    # populate a 1D array of wavelengths (with astropy units of length)
    wavelength = spectra[0]["wavelength"] * u.micron
    rainbow.wavelike["wavelength"] = wavelength * 1

    # populate a 1D array of times (with astropy units of time)
    times = np.arange(len(spectra)) * u.minute
    cheerfully_suggest("The times are totally made up!")
    rainbow.timelike["time"] = times * 1

    # populate a 2D (row = wavelength, col = array of fluxes
    flux = np.zeros((len(wavelength), len(times)))
    uncertainty = np.zeros_like(flux)

    for k in spectra:
        flux[:, k] = spectra[k]["flux"]
        uncertainty[:, k] = spectra[k]["flux_error"]

    rainbow.fluxlike["flux"] = flux * 1
    rainbow.fluxlike["uncertainty"] = uncertainty * 1



from .test_io import *
from .test_basics import *
from .test_operations import *
from .test_normalize import *
from .test_binning import *
from .test_simulations import *
from .test_visualizations import *
from .test_multi import *
from .test_resampling import *
from .test_units import *
from .test_align_wavelengths import *
from .test_shift import *
from .test_conversions import *
from .test_summaries import *
from .test_withmodel import *
from .test_ok import *
from .test_history import *
from .test_spectra import *
from .test_remove_trends import *
from .test_time import *
from .test_fold import *



from ....imports import *
from ..utilities import _add_panel_labels

__all__ = ["plot_with_model", "plot_with_model_and_residuals"]


def plot_with_model(
    self,
    quantity="flux",
    errorbar=True,
    text=True,
    ax=None,
    scaling=1,
    data_plotkw={},
    data_errorbarkw={},
    model_plotkw={},
    minimum_acceptable_ok=1e-10,
    filename=None,
    **kw,
):
    """
    Produce a plot of multiple light curves stacked with an offset,
    with models overplotted (like `plot_lightcurves` with model).

    Parameters
    ----------
    quantity : string
        Which quantity to plot? Options include ['flux', 'residuals'].
    errorbar : boolean
        Should we plot errorbars?
    text : boolean
        Should we label each wavelength?
    data_plotkw : dict
        A dictionary of keywords passed to `plt.plot` for the data.
    data_errorbarkw : dict
        A dictionary of keywords passed to `plt.errorbar` for the data.
    model_plotkw : dict
        A dictionary of keywords passed to `plt.plot` for the model.
    **kw : dict
        All additional keywords will be passed along
        to `.plot_lightcurves. Please see the docstring
        for that function for the full set of options.
    """

    # limit to flux or residuals
    assert quantity in ["flux", "residuals", "scaled_residuals_plus_one"]

    # plot the data
    plotkw = dict(marker="o", linewidth=0, markeredgecolor="none", zorder=0)
    plotkw.update(**data_plotkw)

    errorbarkw = data_errorbarkw
    if quantity == "residuals":
        kw.update(quantity="residuals_plus_one")
    elif quantity == "flux":
        kw.update(quantity="flux")
    ax = self.plot_lightcurves(
        errorbar=errorbar,
        text=text,
        ax=ax,
        plotkw=plotkw,
        errorbarkw=errorbarkw,
        minimum_acceptable_ok=minimum_acceptable_ok,
        scaling=scaling,
        **kw,
    )

    # plot the model
    plotkw = dict(marker=None, linewidth=1, zorder=1)
    plotkw.update(**model_plotkw)
    if quantity == "residuals":
        kw.update(quantity="ones")
    elif quantity == "flux":
        kw.update(quantity="model")
    self.plot_lightcurves(
        errorbar=False,
        text=False,
        plotkw=plotkw,
        minimum_acceptable_ok=minimum_acceptable_ok,
        ax=ax,
        **kw,
    )
    if filename is not None:
        self.savefig(filename)
    return ax


def plot_with_model_and_residuals(
    self,
    figsize=(8, 6),
    filename=None,
    histogram=True,
    histogramkw={},
    residual_scaling=1,
    ax=None,
    label_scatter="{measured*1e6:.0f} ppm [{measured/expected:.1f}x]",
    **kw,
):
    """
    Produce a plot of multiple light curves stacked with an offset,
    with models overplotted (like `plot_lightcurves` with model),
    along with residuals in an adjacent panel.

    Parameters
    ----------
    figsize : tuple
        The figure size for the two-panel side-by-side plot.
    histogram : bool
        Should histograms be plotted for the residuals?
    histogramkw : dict
        Keywords to be passed to plot histograms.
    **kw : dict
        All additional keywords will be passed along
        to `.plot_with_models. Please see the docstring
        for that function for the full set of options.
    """
    if histogram:
        N_columns = 3
        gridspec_kw = dict(width_ratios=[1, 1, 0.1])
    else:
        N_columns = 2
        gridspec_kw = dict()

    if ax is None:
        fi, ax = plt.subplots(
            1,
            N_columns,
            figsize=figsize,
            dpi=300,
            sharey=True,
            # sharex=True,
            constrained_layout=True,
            gridspec_kw=gridspec_kw,
        )

    self.plot_with_model(ax=ax[0], **kw)
    spacing = ax[0]._most_recent_chromatic_plot_spacing
    if "spacing" not in kw:
        kw.update(spacing=spacing)
    if label_scatter is not False:
        kw.update(label_scatter=label_scatter)

    self.plot_with_model(
        quantity="residuals",
        scaling=residual_scaling,
        ax=ax[1],
        **kw,
    )
    if residual_scaling != 1:
        ax[1].set_ylabel(
            r"Residuals $\times$ " + f"{residual_scaling}" + " (+ offsets)"
        )
    else:
        ax[1].set_ylabel("Residuals (+ offsets)")

    if histogram:
        kw = dict(
            orientation="horizontal",
            quantity="residuals",
            expected=True,
            color="auto",
            scaling=residual_scaling,
        )
        kw.update(**histogramkw)
        for i in range(self.nwave):
            self.plot_histogram(i, offset=1 - i * spacing, ax=ax[2], **kw)
        plt.axis("off")

    if filename is not None:
        self.savefig(filename)

    return ax



from ...imports import *
import batman

__all__ = ["inject_transit"]


def inject_transit(self, planet_radius=0.1, **planet_params):

    """
    Simulate a wavelength-dependent planetary transit using
    batman.

    Parameters
    ----------

    planet_radius : float, array
        Two options:
            1D array with same dimensions as wavelength array,
                with each value corresponding to Rp/Rs at that wavelength.
            float representing Rp/Rs if the radius is not wavelength-dependent.
        example value: planet_radius = 0.01,

    planet_params : dict
        All other keywords will be interpreted as other planet parameters.
        Values for planetary parameters for use in batman modelling.
        Any values not supplied will be set to defaults:
            "t0" = time of inferior conjunction (days) (default 0)
            "per" = orbital period (days) (detault 1)
            "a" = semi-major axis (units of stellar radii) (default 15)
            "inc" = inclination (degrees) (default 90)
            "ecc" = eccentricity (default 0)
            "w" = longitude of periastron (degrees)(default 0)
            "limb_dark" = limb-darkening model (default "nonlinear"), possible
                values described in more detail in batman documentation
            "u" = limb-darkening coefficients (default [0.5, 0.1, 0.1, -0.1])
                Can take 3 forms:
                -A single value (if limb-darkening law requires only one value)
                -A 1D list/array of coefficients corresponding to the limb-darkening
                law
                -A 2D array of the form (n_wavelengths, n_coefficients) where
                each row is the set of limb-darkening coefficients corresponding
                to a single wavelength

                Note that this currently does not calculate the appropriate
                coefficient vs wavelength variations itself- there exist codes
                (such as hpparvi/PyLDTk and nespinoza/limb-darkening) which
                can be used for this.

    """

    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("inject_transit", locals())
    h = h.replace("planet_params={", "**{")

    # create a copy of the existing Rainbow
    new = self._create_copy()

    # First, make sure planet_radius has the right dimension.
    if type(planet_radius) != float and len(planet_radius) != self.nwave:
        print(
            "Invalid planet radius array: must be float or have shape "
            + str(np.shape(self.wavelike["wavelength"]))
        )

    # Defaults for planet simulation.
    defaults = {
        "t0": 0,
        "per": 3,
        "a": 10,
        "inc": 90,
        "ecc": 0,
        "w": 0,
        "limb_dark": "nonlinear",
        "u": [0.5, 0.1, 0.1, -0.1],
    }

    # Read in planet parameters.
    for i in range(len(planet_params.keys())):
        key = list(planet_params.keys())[i]
        if key in list(defaults.keys()):
            defaults[key] = planet_params[key]
        else:
            print("Warning: " + str(key) + " not a valid parameter")

    # Initialize batman model.
    params = batman.TransitParams()
    params.t0 = defaults["t0"]
    params.per = defaults["per"]
    params.a = defaults["a"]
    params.inc = defaults["inc"]
    params.ecc = defaults["ecc"]
    params.w = defaults["w"]
    params.limb_dark = defaults["limb_dark"]

    # Deal with limb-darkening.
    if len(np.shape(defaults["u"])) < 2:  # Coefficients constant with wavelength
        u_arr = np.tile(defaults["u"], (self.nwave, 1))

    elif (
        len(np.shape(defaults["u"])) == 2
    ):  # 2D array of coefficients, along wavelength axis
        if np.shape(defaults["u"])[0] != self.nwave:
            print("Shape of limb-darkening array does not match wavelengths.")
            return
        u_arr = defaults["u"]

    else:
        print("Invalid limb-darkening coefficient array.")

    # Read in planetary radius.
    if type(planet_radius) == float:
        rprs = np.zeros(self.nwave) + planet_radius
    else:
        rprs = planet_radius

    planet_flux = np.zeros((self.nwave, self.ntime))

    for i in range(self.nwave):
        params.rp = rprs[i]
        params.u = u_arr[i]
        # print(params.u)
        try:
            m
        except NameError:
            m = batman.TransitModel(params, new.time.to_value("day"))
        planet_flux[i] = m.light_curve(params)

    new.planet_model = planet_flux
    new.flux *= new.planet_model
    new.model = new.fluxlike.get("model", 1) * new.planet_model

    parameters = dict(**defaults)
    parameters.update(rp_unbinned=rprs, u_unbinned=u_arr)
    new.metadata["transit_parameters"] = parameters
    new.metadata["transit_version"] = f"batman-package=={batman.__version__}"
    if np.ndim(u) == 1:
        new.wavelike["injected-u"] = u_arr
    else:
        for i in range(np.shape(u_arr)[1]):
            new.wavelike[f"injected-transit-u{i+1}"] = u_arr[:, i]
    new.wavelike["injected-transit-rp"] = rprs
    # append the history entry to the new Rainbow
    new._record_history_entry(h)

    # return the new object
    return new



from ...imports import *

__all__ = ["plot_spectra"]


def plot_spectra(
    self,
    quantity="flux",
    ax=None,
    spacing=0.1,
    w_unit="micron",
    t_unit="day",
    cmap=None,
    vmin=None,
    vmax=None,
    errorbar=True,
    text=True,
    minimum_acceptable_ok=1,
    scatterkw={},
    errorbarkw={},
    plotkw={},
    textkw={},
    filename=None,
    **kw,
):
    """
    Plot flux as sequence of offset spectrum.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes into which to make this plot.
    spacing : None, float
        The spacing between light curves.
        (Might still change how this works.)
        None uses half the standard dev of entire flux data.
    w_unit : str, astropy.unit.Unit
        The unit for plotting wavelengths.
    t_unit : str, astropy.unit.Unit
        The unit for plotting times.
    cmap : str, matplotlib.colors.Colormap
        The color map to use for expressing wavelength.
    vmin : astropy.units.Quantity
        The minimum value to use for the wavelength colormap.
    vmax : astropy.units.Quantity
        The maximum value to use for the wavelength colormap.
    errorbar : boolean
        Should we plot errorbars?
    text : boolean
        Should we label each spectrum?
    minimum_acceptable_ok : float
        The smallest value of `ok` that will still be included.
        (1 for perfect data, 1e-10 for everything but terrible data, 0 for all data)
    scatterkw : dict
        A dictionary of keywords passed to `plt.scatter`
        so you can have more detailed control over the text
        appearance. Common keyword arguments might include:
        `[alpha, color, s, m, edgecolor, facecolor]` (and more)
        More details are available at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.scatter.html
    errorbarkw : dict
        A dictionary of keywords passed to `plt.errorbar`
        so you can have more detailed control over the plot
        appearance. Common keyword arguments might include:
        `[alpha, elinewidth, color, zorder]` (and more)
        More details are available at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.errorbar.html
    plotkw : dict
        A dictionary of keywords passed to `plt.plot`
        so you can have more detailed control over the plot
        appearance. Common keyword arguments might include:
        `[alpha, clip_on, zorder, marker, markersize,
          linewidth, linestyle, zorder]` (and more)
        More details are available at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.plot.html
    textkw : dict
        A dictionary of keywords passed to `plt.text`
        so you can have more detailed control over the text
        appearance. Common keyword arguments might include:
        `[alpha, backgroundcolor, color, fontfamily, fontsize,
          fontstyle, fontweight, rotation, zorder]` (and more)
        More details are available at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.text.html
    **kw : dict
        Any additional keywords will be stored as `kw`.
        Nothing will happen with them.
    """

    if len(kw) > 0:
        message = f"""
        You provided the keyword argument(s)
        {kw}
        but this function doesn't know how to
        use them. Sorry!
        """
        cheerfully_suggest(message)

    # make sure that the wavelength-based colormap is defined
    self._make_sure_cmap_is_defined(cmap=cmap, vmin=vmin, vmax=vmax)

    w_unit, t_unit = u.Unit(w_unit), u.Unit(t_unit)

    min_wave = np.nanmin(self.wavelength.to_value(w_unit))

    # make sure ax is set up
    if ax is None:
        fi = plt.figure(
            figsize=plt.matplotlib.rcParams["figure.figsize"][::-1],
            constrained_layout=True,
        )
        ax = plt.subplot()
    plt.sca(ax)

    # figure out the spacing to use
    if spacing is None:
        try:
            spacing = ax._most_recent_chromatic_plot_spacing
        except AttributeError:
            spacing = 3 * np.nanstd(self.get(quantity))
    ax._most_recent_chromatic_plot_spacing = spacing

    # TO-DO: check if this Rainbow has been normalized
    '''cheerfully_suggest(
        """
    It's not clear if/how this object has been normalized.
    Be aware that the baseline flux levels may therefore
    be a little bit funny in .plot()."""
    )'''
    with quantity_support():

        #  loop through times
        for i, t in enumerate(self.time):
            # grab the spectrum for this particular time
            w, y, sigma = self.get_ok_data_for_time(
                i, minimum_acceptable_ok=minimum_acceptable_ok, y=quantity
            )
            if np.any(np.isfinite(y)):

                plot_x = w.to_value(w_unit)

                # add an offset to this spectrum
                plot_y = -i * spacing + u.Quantity(y).value
                plot_sigma = u.Quantity(sigma).value

                default_color = "black"

                # set default for background line plot
                this_plotkw = dict(color=default_color, zorder=-1)
                this_plotkw.update(**plotkw)

                # set default for scatter plot with points
                this_scatterkw = dict(
                    marker="o",
                    linestyle="-",
                    c=plot_x,
                    cmap=self.cmap,
                    norm=self.norm,
                )
                this_scatterkw.update(**scatterkw)

                # set default for error bar lines
                this_errorbarkw = dict(
                    color=default_color, linewidth=0, elinewidth=1, zorder=-1
                )
                this_errorbarkw.update(**errorbarkw)

                if errorbar:
                    plt.errorbar(
                        plot_x,
                        plot_y,
                        yerr=plot_sigma,
                        **this_errorbarkw,
                    )
                plt.plot(plot_x, plot_y, **this_plotkw)
                plt.scatter(plot_x, plot_y, **this_scatterkw)

                # add text labels next to each spectrum
                this_textkw = dict(va="center", color=default_color)
                this_textkw.update(**textkw)
                if text:
                    plt.text(
                        min_wave,
                        np.median(plot_y) - 0.5 * spacing,
                        f"{t.to_value(t_unit):.2f} {t_unit.to_string('latex_inline')}",
                        **this_textkw,
                    )

        # add text labels to the plot
        plt.xlabel(f"Wavelength ({w_unit.to_string('latex_inline')})")
        plt.ylabel("Relative Flux (+ offsets)")
        if self.get("wscale") == "log":
            plt.xscale("log")
        plt.title(self.get("title"))
    if filename is not None:
        self.savefig(filename)
    return ax



from ...imports import *

__all__ = ["plot_lightcurves"]


def plot_lightcurves(
    self,
    quantity="flux",
    ax=None,
    spacing=None,
    w_unit="micron",
    t_unit="day",
    cmap=None,
    vmin=None,
    vmax=None,
    errorbar=True,
    text=True,
    minimum_acceptable_ok=1,
    plotkw={},
    errorbarkw={},
    textkw={},
    filename=None,
    scaling=1,
    label_scatter=False,
    **kw,
):
    """
    Plot flux as sequence of offset light curves.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes into which to make this plot.
    spacing : None, float
        The spacing between light curves.
        (Might still change how this works.)
        None uses half the standard dev of entire flux data.
    w_unit : str, astropy.unit.Unit
        The unit for plotting wavelengths.
    t_unit : str, astropy.unit.Unit
        The unit for plotting times.
    cmap : str, matplotlib.colors.Colormap
        The color map to use for expressing wavelength.
    vmin : astropy.units.Quantity
        The minimum value to use for the wavelength colormap.
    vmax : astropy.units.Quantity
        The maximum value to use for the wavelength colormap.
    errorbar : boolean
        Should we plot errorbars?
    text : boolean
        Should we label each lightcurve?
    minimum_acceptable_ok : float
        The smallest value of `ok` that will still be included.
        (1 for perfect data, 1e-10 for everything but terrible data, 0 for all data)
    plotkw : dict
        A dictionary of keywords passed to `plt.plot`
        so you can have more detailed control over the plot
        appearance. Common keyword arguments might include:
        `[alpha, clip_on, zorder, marker, markersize,
          linewidth, linestyle, zorder]` (and more)
        More details are available at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.plot.html
    errorbarkw : dict
        A dictionary of keywords passed to `plt.errorbar`
        so you can have more detailed control over the plot
        appearance. Common keyword arguments might include:
        `[alpha, elinewidth, color, zorder]` (and more)
        More details are available at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.errorbar.html
    textkw : dict
        A dictionary of keywords passed to `plt.text`
        so you can have more detailed control over the text
        appearance. Common keyword arguments might include:
        `[alpha, backgroundcolor, color, fontfamily, fontsize,
          fontstyle, fontweight, rotation, zorder]` (and more)
        More details are available at
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.text.html
    kw : dict
        Any additional keywords will be stored as `kw`.
        Nothing will happen with them.
    """
    if len(kw) > 0:
        message = f"""
        You provided the keyword argument(s)
        {kw}
        but this function doesn't know how to
        use them. Sorry!
        """
        cheerfully_suggest(message)

    # make sure that the wavelength-based colormap is defined
    self._make_sure_cmap_is_defined(cmap=cmap, vmin=vmin, vmax=vmax)

    w_unit, t_unit = u.Unit(w_unit), u.Unit(t_unit)

    min_time = np.nanmin(self.time.to_value(t_unit))
    max_time = np.nanmax(self.time.to_value(t_unit))

    # make sure ax is set up
    if ax is None:
        fi = plt.figure(
            figsize=plt.matplotlib.rcParams["figure.figsize"][::-1],
            constrained_layout=True,
        )
        ax = plt.subplot()
    plt.sca(ax)

    # figure out the spacing to use
    if spacing is None:
        try:
            spacing = ax._most_recent_chromatic_plot_spacing
        except AttributeError:
            spacing = 3 * np.nanstd(self.get(quantity))
    ax._most_recent_chromatic_plot_spacing = spacing

    # TO-DO: check if this Rainbow has been normalized
    if self._is_probably_normalized() or "model" in self.fluxlike:
        label_y = "1 - (0.5 + i) * spacing"
        ylim = 1 - np.array([self.nwave + 1, -1]) * spacing
    else:
        label_y = "np.median(plot_y) - 0.5 * spacing"
        cheerfully_suggest(
            """
            It's not clear if/how this object has been normalized.
            Be aware that the baseline flux levels may therefore
            be a little bit funny in .plot()."""
        )
        ylim = None
    with quantity_support():

        if label_scatter:
            measured_rms = self.get_measured_scatter(quantity="residuals")
            expected_rms = self.get_expected_uncertainty()

        #  loop through wavelengths
        for i, w in enumerate(self.wavelength):

            # grab the quantity and yerr for this particular wavelength
            t, y, sigma = self.get_ok_data_for_wavelength(
                i, minimum_acceptable_ok=minimum_acceptable_ok, y=quantity
            )

            if np.any(np.isfinite(y)):

                plot_x = t.to_value(t_unit)

                # add an offset to this quantity
                plot_y = -i * spacing + (u.Quantity(y).value - 1) * scaling + 1
                plot_sigma = u.Quantity(sigma).value * scaling

                # get the color for this quantity
                color = self.get_wavelength_color(w)

                # plot the data points (with offsets)
                this_plotkw = dict(marker="o", linestyle="-", color=color)
                this_plotkw.update(**plotkw)

                # set default for error bar lines
                this_errorbarkw = dict(
                    color=color, linewidth=0, elinewidth=1, zorder=-1
                )
                this_errorbarkw.update(**errorbarkw)

                if errorbar:
                    plt.errorbar(
                        plot_x,
                        plot_y,
                        yerr=plot_sigma,
                        **this_errorbarkw,
                    )
                plt.plot(plot_x, plot_y, **this_plotkw)

                # add text labels next to each quantity plot
                this_textkw = dict(va="center", color=color)
                this_textkw.update(**textkw)
                if text:
                    plt.text(
                        min_time,
                        eval(label_y),
                        f"{w.to_value(w_unit):.2f} {w_unit.to_string('latex_inline')}",
                        **this_textkw,
                    )

                if label_scatter is not False:
                    this_textkw.update(ha="right")
                    measured = measured_rms[i]
                    expected = expected_rms[i]
                    cadence = self.dt
                    if text:
                        plt.text(
                            max_time,
                            eval(label_y),
                            eval(f'f"{label_scatter}"'),
                            **this_textkw,
                        )

        # add text labels to the plot
        plt.xlabel(f"{self._time_label} ({t_unit.to_string('latex_inline')})")
        plt.ylabel("Relative Flux (+ offsets)")
        if ylim is not None:
            if ylim[1] != ylim[0]:
                plt.ylim(*ylim)
        plt.title(self.get("title"))

    if filename is not None:
        self.savefig(filename)
    return ax



from ...imports import *

__all__ = ["inflate_uncertainty"]


def inflate_uncertainty(
    self,
    minimum_inflate_ratio = 1.0,
    **kw,
):  
    """
    A quick tool to inflate uncertainties to photon noise.

    Parameters
    ----------
    method : str
        What method should be used to make an approximate model
        for smooth trends that will then be subtracted off?

        `differences` will do an extremely rough filtering
        of replacing the fluxes with their first differences.
        Trends that are smooth relative to the noise will
        be removed this way, but sharp features will remain.
        Required keywords:
            None.

        `median_filter` is a wrapper for scipy.signal.median_filter.
        It smoothes each data point to the median of its surrounding
        points in time and/or wavelength. Required keywords:
            `size` = centered on each point, what shape rectangle
            should be used to select surrounding points for median?
            The dimensions are (nwavelengths, ntimes), so `size=(3,7)`
            means we'll take the median across three wavelengths and
            seven times. Default is `(1,5)`.

        `savgol_filter` is a wrapper for scipy.signal.savgol_filter.
        It applies a Savitzky-Golay filter for polynomial smoothing.
        Required keywords:
            `window_length` = the length of the filter window,
            which must be a positive odd integer. Default is `5`.
            `polyorder` = the order of the polynomial to use.
            Default is `2`.

        `polyfit` is a wrapper for numpy.polyfit to use a weighted
        linear least squares polynomial fit to remove smooth trends
        in time. Required keywods:
            `deg` = the polynomial degree, which must be a positive
            integer. Default is `1`, meaning a line.

        `custom` allow users to pass any fluxlike array of model
        values for an astrophysical signal to remove it. Required
        keywords:
            `model` = the (nwavelengths, ntimes) model array

    minimum_inflate_ratio : float
	the minimum inflate_ratio that can be. We don't want people 
	to deflate uncertainty unless a very specific case of unstable 
	pipeline output.

    kw : dict
        Any additional keywords will be passed to the function
        that does the filtering. See `method` keyword for options.

    Returns
    -------
    removed : Rainbow
        The Rainbow with estimated signals removed.
    """

    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("inflate_uncertainty", locals())

    # TODO, think about more careful treatment of uncertainties + good/bad data
    new = self._create_copy()
	    
    trend_removed = new.remove_trends(**kw)
    measured_scatter = trend_removed.get_measured_scatter(method='MAD',minimum_acceptable_ok=1e-10)

    expected_uncertainty = trend_removed.get_expected_uncertainty()

    inflate_ratio = measured_scatter/expected_uncertainty
    new.wavelike['inflate_ratio'] = inflate_ratio
    if np.min(inflate_ratio) < minimum_inflate_ratio:
    	raise Exception(f"One or more inflate ratios are below the minimum_inflate_ratio = {minimum_inflate_ratio}, if you really want to proceed please consider lower minimum_inflate_ratio")

    new.uncertainty = new.uncertainty*inflate_ratio[:,np.newaxis]  
  
    # append the history entry to the new Rainbow
    new._record_history_entry(h)

    # return the new Rainbow
    return new



from .withmodel import *
from .rainbow import *
from .simulated import *
from .multi import *


def read_rainbow(filepath, **kw):
    """
    A friendly wrapper to load time-series spectra and/or
    multiwavelength light curves into a `chromatic` Rainbow
    object. It will try its best to pick the best reader
    and return the most useful kind of object.

    🦋🌅2️⃣🪜🎬👀🇮🇹📕🧑‍🏫🌈

    Parameters
    ----------
    filepath : str, list
        The file or files to open.
    **kw : dict
        All other keyword arguments will be passed to
        the `Rainbow` initialization.

    Returns
    -------
    rainbow : Rainbow, RainbowWithModel
        The loaded data!
    """
    r = Rainbow(filepath, **kw)
    if "model" in r.fluxlike:
        return RainbowWithModel(**r._get_core_dictionaries())
    else:
        return r



"""
Define a reader for NIRISS/SOSS extract1d files produced using the ATOCA
algorithm.
"""
from ...imports import *

__all__ = ["from_atoca"]


def get_time_axis(header):
    """Define the time axis for a NIRISS/SOSS TSO

    Parameters
    ----------
    header : astropy header object
        Header of the primary HDU.

    Returns
    -------
    t : array-like
        TSO time axis in jd.
    nint : int
        Number of integrations in the TSO.
    """

    # Each integration is not time stamped - need to reconstruct the time axis
    # based on available information.
    # Get the observation start time
    t_start = header["DATE-OBS"] + "T" + header["TIME-OBS"]
    # Get frame time and convert to days.
    tframe = header["TFRAME"] * u.s.to(u.d)
    # Get number of frames, groups and integrations.
    # nframe willl generally be 1 for NIRISS/SOSS, but just in case...
    nframe = header["NFRAMES"]
    ngroup = header["NGROUPS"] + 1  # Need to add one for reset.
    nint = header["NINTS"]
    # Construct time axis and give units of days.
    t_start = Time(t_start, format="isot", scale="utc")
    t = np.arange(nint) * tframe * nframe * ngroup + t_start.jd
    t *= u.d

    return t, nint


def from_atoca(rainbow, filepath, order=1):
    """Initialize a Rainbow based on the outputs of the ATOCA 1D extraction
    algorithm. This will only be used for NIRISS/SOSS spectra.

    Parameters
    ----------
    rainbow : Rainbow
        The object to be populated. (This is intended to be used as a class
        method of Rainbow or of a class derived from Rainbow, as a way of
        initializing an object from files.)
    filepath : str
        The path to the file, or a glob-friendly string pointing to a group of
        files that should all be loaded together (for example, if an exposure
        was split into multiple segments).
    order : int
        Diffraction order to be unpacked, either 1 or 2. Currently both cannot
        be unpacked simultaneously.
    """

    # Verify that the requested order makes sense.
    if order not in [1, 2]:
        msg = "Only orders 1 and 2 are extracted by ATOCA."
        raise NotImplementedError(msg)
    else:
        # Ensure that the user knows which order they are getting.
        other_order = {1: 2, 2: 1}[order]
        cheerfully_suggest(
            f"""
        You are loading NIRISS order {order} from the ATOCA file
        {filepath}
        If you want the other order, try providing the
        `order={other_order}` as a keyword argument to `Rainbow()`.
        """
        )

    filenames = expand_filenames(filepath)

    # Loop over all input filenames.
    for i_file, f in enumerate(tqdm(filenames, leave=False)):
        hdu_list = fits.open(f)

        # Useful initializations for later.
        quantities = {}
        first_time = True

        # For the first file, define common time axis.
        if i_file == 0:
            # Create time axis
            times, nints = get_time_axis(hdu_list["PRIMARY"].header)
            rainbow.timelike["time"] = times * 1

        # Loop over all extensions.
        for i in range(1, len(hdu_list)):
            # Only consider extract1d extensions of the correct order.
            if hdu_list[i].header["EXTNAME"] != "EXTRACT1D":
                continue
            if hdu_list[i].header["SPORDER"] != order:
                continue
            # Unpack each of the data types stored in each extension.
            # This includes the usual wavelength, flux, and error, but also
            # many other things which may or may not be useful (or even
            # populated). But keep them just in case.
            for quantity in hdu_list[i].data.names:
                if first_time:
                    quantities[quantity] = hdu_list[i].data[quantity]
                else:
                    quantities[quantity] = np.vstack(
                        [quantities[quantity], hdu_list[i].data[quantity]]
                    )
            first_time = False

    # Pack all the above data into a Rainbow object.
    for quantity in quantities.keys():
        # Try to keep the main key names consistant with other formats.
        # Flux-like things in electron/s and wavelengths in microns
        if quantity == "FLUX":
            rainbow.fluxlike["flux"] = quantities[quantity].T * 1.6 * u.electron / u.s
        elif quantity == "FLUX_ERROR":
            rainbow.fluxlike["uncertainty"] = (
                quantities[quantity].T * 1.6 * u.electron / u.s
            )
        elif quantity == "WAVELENGTH":
            rainbow.wavelike["wavelength"] = (
                np.nanmedian(quantities[quantity], axis=0) * u.micron * 1
            )
        else:
            rainbow.fluxlike[quantity] = quantities[quantity].T * 1

    # Warn user if the number of unpacked integrations doesn't match the
    # expected amount.
    n_filled_times = rainbow.fluxlike["flux"].shape[1]
    if n_filled_times != nints:
        cheerfully_suggest(
            f"""The extract1d header(s) indicate there should be
            {rainbow.ntime} integrations, but only {n_filled_times} columns of
            the flux array were populated. Are you perhaps missing some
            segment files?"""
        )



from ...imports import *

__all__ = ["from_text"]


def from_text(rainbow, filename, **kwargs):
    """
    Populate a Rainbow from a generic text file.
    This text file must at least contain the columns
    `wavelength`, `time`, `flux`, `uncertainty`

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated. (This is intended
        to be used as a class method of Rainbow or
        of a class derived from Rainbow, as a way of
        initializing an object from files.)
    filename : str
        The path to the file.
    """

    # load the data
    data = ascii.read(filename)

    # pull out some variables
    t = sorted(np.unique(data["time"]))
    w = sorted(np.unique(data["wavelength"]))

    fluxlike = {}
    fluxlike_keys = list(data.colnames)
    fluxlike_keys.remove("time")
    fluxlike_keys.remove("wavelength")
    for k in fluxlike_keys:
        fluxlike[k] = np.ones(shape=(len(w), len(t)))
    for i in tqdm(range(len(data)), leave=False):
        # this is slow, but general
        i_time = t == data[i]["time"]
        i_wavelength = w == data[i]["wavelength"]
        for k in fluxlike_keys:
            fluxlike[k][i_wavelength, i_time] = data[i][k]

    # do a slightly better guess of the uncertainty column
    if "uncertainty" not in fluxlike:
        for k in ["error", "flux_error", "sigma", "unc"]:
            try:
                fluxlike["uncertainty"] = fluxlike[k] * 1
                break
            except KeyError:
                pass

    timelike = {}
    try:
        t.unit
    except AttributeError:
        t *= u.day
    timelike["time"] = t

    wavelike = {}
    try:
        w.unit
    except AttributeError:
        w *= u.micron
    wavelike["wavelength"] = w

    # populate the rainbow
    rainbow._initialize_from_dictionaries(
        wavelike=wavelike, timelike=timelike, fluxlike=fluxlike
    )



from ..rainbows import *
from .setup_tests import *


def test_fold(N=5):

    for i in range(5):
        original_time = (
            np.linspace(np.random.uniform(-0.5, -0.2), np.random.uniform(0.2, 0.5), 500)
            * u.day
        )
        period = np.random.uniform(1, 10) * u.day
        t0 = np.random.uniform(-10, 10) * u.day
        N = np.random.randint(-10, 10)
        s = (
            SimulatedRainbow(time=original_time + t0 + period * N, R=5)
            .inject_transit(per=period.to_value("day"), t0=t0.to_value("day"))
            .inject_noise(signal_to_noise=1000)
        )
        f = s.fold(period=period, t0=t0)
        assert np.all(np.isclose(f.time, original_time, atol=1e-12))
    fi, ax = plt.subplots(1, 2, figsize=(8, 3), constrained_layout=True)
    s.imshow(ax=ax[0])
    f.imshow(ax=ax[1])
    plt.savefig(
        os.path.join(test_directory, "demonstration-of-folding-to-period-and-t0.pdf")
    )



from ..rainbows import *
from ..rainbows.actions.align_wavelengths import _create_shared_wavelength_axis
from .setup_tests import *


def create_simulation_with_wobbly_wavelengths(
    fractional_shift=0.002,
    signal_to_noise=100,
    dw=0.001 * u.micron,
    wlim=[0.95, 1.05] * u.micron,
    dt=30 * u.minute,
    **kw
):

    # set up a function to create a fake absorption line spectrum
    N = 10
    centers = (
        np.random.uniform(wlim[0].to_value("micron"), wlim[1].to_value("micron"), N)
        * u.micron
    )
    sigmas = np.random.uniform(0.001, 0.003, N) * u.micron
    amplitudes = np.random.uniform(0, 0.5, N)

    def fake_spectrum(w):
        f = u.Quantity(np.ones(np.shape(w)))
        for c, s, a in zip(centers, sigmas, amplitudes):
            f *= 1 - a * np.exp(-0.5 * (w - c) ** 2 / s**2)
        return f

    # create
    r = SimulatedRainbow(dw=dw, wlim=wlim, dt=dt, **kw).inject_noise(
        signal_to_noise=signal_to_noise
    )
    wobbly_wavelengths = r.wavelength[:, np.newaxis] * np.random.normal(
        1, fractional_shift, r.ntime
    )
    r.fluxlike["wavelength_2d"] = wobbly_wavelengths
    r.fluxlike["model"] = fake_spectrum(r.fluxlike["wavelength_2d"])
    r.fluxlike["flux"] = r.fluxlike["flux"] * r.fluxlike["model"]

    return r


def test_create_shared_wavelength_axis(fractional_shift=0.002, dw=0.01 * u.micron):
    r = create_simulation_with_wobbly_wavelengths(
        fractional_shift=fractional_shift, dw=dw
    )
    _create_shared_wavelength_axis(r, wscale="linear", visualize=True)
    plt.savefig(
        os.path.join(
            test_directory,
            "demonstration-of-creating-shared-wavelength-axis.pdf",
        )
    )


def test_align_wavelengths(fractional_shift=0.002, dw=0.001 * u.micron, **kw):
    r = create_simulation_with_wobbly_wavelengths(
        fractional_shift=fractional_shift, dw=dw, **kw
    )
    a = r.align_wavelengths()
    plt.close("all")
    fi, ax = plt.subplots(2, 2, dpi=300, figsize=(8, 6), constrained_layout=True)
    for i, x in enumerate([r, a]):
        plt.sca(ax[0, i])
        plt.imshow(x.flux, aspect="auto", cmap="gray")
        plt.title(["original", "aligned"][i] + " flux")
        plt.xlabel("time index")
        plt.ylabel("wavelength index")
        plt.colorbar()

        plt.sca(ax[1, i])
        plt.imshow(x.fluxlike["wavelength_2d"], aspect="auto")
        plt.title(["original", "aligned"][i] + " wavelength")
        plt.xlabel("time index")
        plt.ylabel("wavelength index")
        plt.colorbar()

    plt.savefig(
        os.path.join(
            test_directory,
            "demonstration-of-aligning-2D-wavelengths-to-shared-axis.pdf",
        )
    )


def test_align_wavelengths_with_not_ok_data(visualize=False):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for ok_fraction in [0.25, 0.5, 0.75, 1.0]:
            r = create_simulation_with_wobbly_wavelengths(
                fractional_shift=0.0005,
                wlim=[0.99, 1.01] * u.micron,
                dw=0.001 * u.micron,
                dt=20 * u.minute,
            )
            r.fluxlike["ok"] = np.random.uniform(size=r.shape) < ok_fraction
            cautious = r.align_wavelengths(minimum_acceptable_ok=1)
            carefree = r.align_wavelengths(minimum_acceptable_ok=0)
            if visualize:
                cautious.imshow_quantities()
                plt.suptitle(ok_fraction)
                carefree.imshow_quantities()
                plt.suptitle(ok_fraction)

        # if np.any(r.ok == 0):
        #    assert np.any((carefree.ok != 1) & (carefree.ok != 0))



# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["from_eureka_channels", "from_eureka_S5"]


def from_eureka_channels(rainbow, filepath):
    """
    Populate a Rainbow from Eureka! S5 modeling outputs,
    which are wavelength-binned light curves with models.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.

    filepath : str
        The path to the files to load.

    """
    # get list of files
    filenames = expand_filenames(filepath)

    # load them into tables
    tables = [ascii.read(f) for f in filenames]

    # load the time and wavelength arrays
    rainbow.timelike["time"] = tables[0]["time"] * u.day
    rainbow.wavelike["wavelength"] = [t["wavelength"][0] for t in tables] * u.micron
    rainbow.wavelike["wavelength_lower"] = [
        (t["wavelength"] - t["bin_width"])[0] for t in tables
    ] * u.micron
    rainbow.wavelike["wavelength_upper"] = [
        (t["wavelength"] + t["bin_width"])[0] for t in tables
    ] * u.micron
    keys_used = ["time", "wavelength", "bin_width"]

    # populate the fluxlike arrays
    for k in ["lcdata", "lcerr", "model", "residuals"]:
        rainbow.fluxlike[k] = np.array([t[k] for t in tables])
        keys_used.append(k)

    for k in tables[0].colnames:
        if k != keys_used:
            rainbow.fluxlike[k] = np.array([t[k] for t in tables])

    rainbow.fluxlike["flux"] = rainbow.lcdata * 1
    rainbow.fluxlike["uncertainty"] = rainbow.lcerr * 1


from_eureka_S5 = from_eureka_channels



# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["from_eureka_LCData", "from_eureka_S4"]


def from_eureka_LCData(rainbow, filepath):
    """
    Populate a Rainbow from Eureka! S4 LC outputs,
    which are wavelength-binned light curves.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.

    filepath : str
        The path to the file to load.

    """

    try:
        import astraeus.xarrayIO as xrio
    except ModuleNotFoundError:
        cheerfully_suggest(
            f"""
        You are trying import `astraeus`, which is needed for
        reading and writing `Eureka!` pipeline files. We're
        having trouble importing it, so you could please
        confirm that it is installed into you current
        environment by running the following command?

        pip install git+https://github.com/kevin218/Astraeus.git

        Once you've installed it, please restart whatever
        code you were trying to run, and hopefully it
        will work! Thanks for your patience!
        """
        )
        return None

    # use Astraeus to read in the file
    dataset = xrio.readXR(filepath, verbose=False)

    keys_used = []

    # populate a 1D array of wavelengths (with astropy units of length)
    k = "wavelength"
    w_without_unit = dataset[k].data
    w_unit = u.Unit(dataset[k].attrs["wave_units"])
    rainbow.wavelike["wavelength"] = w_without_unit * w_unit
    keys_used.append(k)

    for k_theirs, k_ours in zip(
        ["wave_hi", "wave_low"], ["wavelength_lower", "wavelength_lower"]
    ):
        w_without_unit = dataset[k_theirs].data
        rainbow.wavelike[k_ours] = w_without_unit * w_unit
        keys_used.append(k_theirs)

    # populate a 1D array of times (with astropy units of time)
    k = "time"
    t_without_unit = dataset[k].data
    if dataset[k].attrs["time_units"] == "BJD_TDB":
        t_unit = u.day
    else:
        t_unit = u.Unit(dataset[t_key].attrs["time_units"])
    rainbow.timelike["time"] = t_without_unit * t_unit
    keys_used.append(k)

    # populate a 2D (row = wavelength, col = time) array of fluxes
    for k in dataset.data_vars.keys():
        shape = dataset[k].shape
        if shape == rainbow.shape:
            q_without_unit = dataset[k].data
            for k_unit in ["flux_units"]:
                try:
                    q_unit = u.Unit(dataset[k_unit])
                    continue
                except (KeyError, ValueError):
                    q_unit = u.Unit()
            rainbow.fluxlike[k] = q_without_unit * q_unit
            keys_used.append(k)
        test_shapes = (rainbow.shape, rainbow.wavelength.shape, rainbow.time.shape)

    rainbow.fluxlike["flux"] = rainbow.data * 1
    rainbow.fluxlike["uncertainty"] = rainbow.err * 1

    keys_used.extend(["wave_mid", "wave_err"])
    for k in dataset.keys():
        if (shape in test_shapes) or (shape[::-1] in test_shapes):
            if k not in keys_used:
                cheerfully_suggest(
                    f"""
                While reading a Eureka S4 LCData dataset, the key '{k}'
                has not been stored as `fluxlike`, `wavelike`, or `timelike`,
                yet its shape {shape} feels like it might match one
                of those. If you'd like to read in that quantity,
                please submit an Issue to the `chromatic` github
                to make sure it gets added!
                """
                )


from_eureka_S4 = from_eureka_LCData



from ...imports import *


def trim_times(self, just_edges=True, when_to_give_up=1, minimum_acceptable_ok=1):
    """
    Trim times that are all (or mostly) not numbers.

    Parameters
    ----------
    just_edges : bool
        Should we only trim the outermost bad time bins?
            `True` = Just trim off the bad edges and keep
            interior bad values. Keeping interior data, even if
            they're all bad, often helps to make for more
            intuititive imshow plots.
            `False` = Trim off every bad time, whether it's on
            the edge or somewhere in the middle of the dataset.
            The resulting Rainbow will be smaller, but it might
            be a little tricky to visualize with imshow.
    when_to_give_up : float
        The fraction of wavelengths that must be nan or not OK
        for the entire time to be considered bad (default = 1).
            `1.0` = trim only if all wavelengths are bad
            `0.5` = trim if more than 50% of wavelengths are bad
            `0.0` = trim if any wavelengths are bad
    minimum_acceptable_ok : float
        The numbers in the `.ok` attribute express "how OK?" each
        data point is, ranging from 0 (not OK) to 1 (super OK).
        In most cases, `.ok` will be binary, but there may be times
        where it's intermediate (for example, if a bin was created
        from some data that were not OK and some that were).
        The `minimum_acceptable_ok` parameter allows you to specify what
        level of OK-ness for a point to not get trimmed.
    """

    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("trim_times", locals())

    # figure out which times should be considered bad
    is_nan = np.isnan(self.flux)
    isnt_ok = self.ok < minimum_acceptable_ok
    fraction_bad = np.sum(is_nan | isnt_ok, axis=self.waveaxis) / self.nwave
    should_be_kept = fraction_bad < when_to_give_up

    # only make cuts on the edges (if desired)
    if just_edges:
        isnt_before_first = np.cumsum(should_be_kept) > 0
        isnt_after_last = (np.cumsum(should_be_kept[::-1]) > 0)[::-1]
        isnt_edge = isnt_before_first & isnt_after_last
        should_be_kept = should_be_kept | isnt_edge

    # actually try the Rainbow
    new = self[:, should_be_kept]
    new._remove_last_history_entry()

    # append the history entry to the new Rainbow
    new._record_history_entry(h)

    # return the new Rainbow
    return new


def trim_wavelengths(self, just_edges=True, when_to_give_up=1, minimum_acceptable_ok=1):
    """
    Trim wavelengths that are all (or mostly) not numbers.

    Parameters
    ----------
    just_edges : bool
        Should we only trim the outermost bad wavelength bins?
            `True` = Just trim off the bad edges and keep
            interior bad values. Keeping interior data, even if
            they're all bad, often helps to make for more
            intuititive imshow plots.
            `False` = Trim off every bad wavelength, whether it's on
            the edge or somewhere in the middle of the dataset.
            The resulting Rainbow will be smaller, but it might
            be a little tricky to visualize with imshow.
    when_to_give_up : float
        The fraction of times that must be nan or not OK
        for the entire wavelength to be considered bad (default = 1).
            `1.0` = trim only if all times are bad
            `0.5` = trim if more than 50% of times are bad
            `0.0` = trim if any times are bad
    minimum_acceptable_ok : float
        The numbers in the `.ok` attribute express "how OK?" each
        data point is, ranging from 0 (not OK) to 1 (super OK).
        In most cases, `.ok` will be binary, but there may be times
        where it's intermediate (for example, if a bin was created
        from some data that were not OK and some that were).
        The `minimum_acceptable_ok` parameter allows you to specify what
        level of OK-ness for a point to not get trimmed.
    """
    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("trim_wavelengths", locals())

    # figure out which wavelengths should be considered bad
    is_nan = np.isnan(self.flux)
    isnt_ok = self.ok < minimum_acceptable_ok
    fraction_bad = np.sum(is_nan | isnt_ok, axis=self.timeaxis) / self.ntime
    should_be_kept = fraction_bad < when_to_give_up

    # only make cuts on the edges (if desired)
    if just_edges:
        isnt_before_first = np.cumsum(should_be_kept) > 0
        isnt_after_last = (np.cumsum(should_be_kept[::-1]) > 0)[::-1]
        isnt_edge = isnt_before_first & isnt_after_last
        should_be_kept = should_be_kept | isnt_edge

    # actually try the Rainbow
    new = self[should_be_kept, :]
    new._remove_last_history_entry()

    # append the history entry to the new Rainbow
    new._record_history_entry(h)

    # return the new Rainbow
    return new


def trim(self, just_edges=True, when_to_give_up=1, minimum_acceptable_ok=1):
    """
    just_edges : bool
        Should we only trim the outermost bad wavelength bins?
            `True` = Just trim off the bad edges and keep
            interior bad values. Keeping interior data, even if
            they're all bad, often helps to make for more
            intuititive imshow plots.
            `False` = Trim off every bad wavelength, whether it's on
            the edge or somewhere in the middle of the dataset.
            The resulting Rainbow will be smaller, but it might
            be a little tricky to visualize with imshow.
    when_to_give_up : float
        The fraction of times that must be nan or not OK
        for the entire wavelength to be considered bad (default = 1).
            `1.0` = trim only if all times are bad
            `0.5` = trim if more than 50% of times are bad
            `0.0` = trim if any times are bad
    minimum_acceptable_ok : float
        The numbers in the `.ok` attribute express "how OK?" each
        data point is, ranging from 0 (not OK) to 1 (super OK).
        In most cases, `.ok` will be binary, but there may be times
        where it's intermediate (for example, if a bin was created
        from some data that were not OK and some that were).
        The `minimum_acceptable_ok` parameter allows you to specify what
        level of OK-ness for a point to not get trimmed.
    """

    trimmed = self.trim_times(
        when_to_give_up=when_to_give_up,
        just_edges=just_edges,
        minimum_acceptable_ok=minimum_acceptable_ok,
    )
    trimmed = trimmed.trim_wavelengths(
        when_to_give_up=when_to_give_up,
        just_edges=just_edges,
        minimum_acceptable_ok=minimum_acceptable_ok,
    )

    return trimmed



# import the general list of packages
from ...imports import *

# define list of the only things that will show up in imports
__all__ = ["from_eureka_SpecData", "from_eureka_S3"]


def from_eureka_SpecData(rainbow, filepath, optimal=True):
    """
    Populate a Rainbow from Eureka! S3/S4 SpecData outputs,
    which are time-series extracted spectra.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.

    filepath : str
        The path to the file to load.

    optimal : bool
        Should the optimal extraction be used as the flux?
        If True, (flux, uncertainty) = (optspec, opterr)
        If False, (flux, uncertainty) = (stdspec, sqrt(stdvar))
    """

    try:
        import astraeus.xarrayIO as xrio
    except ModuleNotFoundError:
        cheerfully_suggest(
            f"""
        You are trying import `astraeus`, which is needed for
        reading and writing `Eureka!` pipeline files. We're
        having trouble importing it, so you could please
        confirm that it is installed into you current
        environment by running the following command?

        pip install git+https://github.com/kevin218/Astraeus.git

        Once you've installed it, please restart whatever
        code you were trying to run, and hopefully it
        will work! Thanks for your patience!
        """
        )
        return None

    # use Astraeus to read in the file
    dataset = xrio.readXR(filepath, verbose=False)

    keys_used = []

    # populate a 1D array of wavelengths (with astropy units of length)
    k = "wave_1d"
    w_without_unit = dataset[k].data
    w_unit = u.Unit(dataset[k].attrs["wave_units"])
    rainbow.wavelike["wavelength"] = w_without_unit * w_unit
    keys_used.append(k)

    # populate a 1D array of times (with astropy units of time)
    k = "time"
    if dataset[k].attrs["time_units"] == "BJD_TDB":
        astropy_times = Time(dataset[k].data, format="mjd", scale="tdb")
        rainbow.set_times_from_astropy(astropy_times, is_barycentric=True)
    else:
        t_unit = u.Unit(dataset[t_key].attrs["time_units"])
        t_without_unit = dataset[k].data
        rainbow.timelike["time"] = t_without_unit * t_unit
    keys_used.append(k)

    # populate a 2D (row = wavelength, col = time) array of fluxes
    for k in dataset.data_vars.keys():
        shape = dataset[k].shape
        if shape[::-1] == rainbow.shape:
            q_without_unit = dataset[k].data.T
            for k_unit in ["flux_units"]:
                try:
                    q_unit = u.Unit(dataset[k_unit])
                    continue
                except (KeyError, ValueError):
                    q_unit = u.Unit()
            rainbow.fluxlike[k] = q_without_unit * q_unit
            keys_used.append(k)
        test_shapes = (rainbow.shape, rainbow.wavelength.shape, rainbow.time.shape)

    if optimal:
        rainbow.fluxlike["flux"] = rainbow.optspec * 1
        rainbow.fluxlike["uncertainty"] = rainbow.opterr * 1
    else:
        rainbow.fluxlike["flux"] = rainbow.stdspec * 1
        rainbow.fluxlike["uncertainty"] = np.sqrt(rainbow.stdvar) * 1

    for k in dataset.keys():
        if k not in keys_used:
            shape = dataset[k].shape
            values = dataset[k].data
            try:
                if shape in test_shapes:
                    rainbow._put_array_in_right_dictionary(k, values)
                elif shape[::-1] in test_shapes:
                    rainbow._put_array_in_right_dictionary(k, values.T)
            except ValueError:
                cheerfully_suggest(
                    f"""
                While reading a Eureka S3 SpecData dataset, the key '{k}'
                has not been stored as `fluxlike`, `wavelike`, or `timelike`,
                but its shape of {shape} feels like maybe it should fit.
                If you'd like to read in that quantity, please submit an
                Issue to the `chromatic` github to make sure it gets added!
                """
                )


from_eureka_S3 = from_eureka_SpecData



from ..rainbows import *
from .setup_tests import *


def test_ok_rows_and_columns():
    s = SimulatedRainbow().inject_noise()
    s.wavelike["ok"] = np.arange(s.nwave) > 10
    s.timelike["ok"] = np.arange(s.ntime) > 5
    s.fluxlike["ok"] = np.ones(s.shape)
    assert s.ok[0, 0] == 0
    assert s.ok[-1, -1] == 1
    assert np.all(s.ok[:10, :] == 0)
    assert np.all(s.ok[:, :5] == 0)


def test_bin_with_not_ok_data():
    for ok_fraction in [0, 0.5, 0.99, 1]:
        a = (
            SimulatedRainbow(dt=2 * u.minute, dw=0.2 * u.micron)
            .inject_transit()
            .inject_noise()
        )
        a.ok = np.random.uniform(size=a.shape) < ok_fraction

        if ok_fraction == 0:
            with pytest.raises((RuntimeError, IndexError)):
                should_fail = a.bin(
                    dw=0.7 * u.micron, dt=20 * u.minute, minimum_acceptable_ok=1
                )
            continue

        cautious = a.bin(dw=0.4 * u.micron, dt=4 * u.minute, minimum_acceptable_ok=1)
        carefree = a.bin(
            dw=0.4 * u.micron, dt=4 * u.minute, minimum_acceptable_ok=0, trim=False
        )
        if np.any(a.ok == 0):
            assert np.any((carefree.ok != 1) & (carefree.ok != 0))


def test_get_helpers():
    s = SimulatedRainbow()
    assert s.get("flux") is s.flux

    s.get_for_wavelength(0).shape == (s.ntime,)
    s.get_for_wavelength(0, "flux").shape == (s.ntime,)
    s.get_for_wavelength(0, "time").shape == (s.ntime,)
    with pytest.raises(RuntimeError):
        s.get_for_wavelength(0, "wavelength")

    s.get_for_time(0).shape == (s.nwave,)
    s.get_for_time(0, "flux").shape == (s.nwave,)
    s.get_for_time(0, "wavelength").shape == (s.nwave,)
    with pytest.raises(RuntimeError):
        s.get_for_time(0, "time")


def test_get_ok_data_helpers(quantity="flux"):
    s = SimulatedRainbow(dw=0.5 * u.micron, dt=20 * u.minute).inject_noise()
    s.ok = np.random.uniform(0, 1, s.shape) > 0.5
    s.imshow()

    fi, ax = plt.subplots(2, 2, sharex="col", sharey=True, constrained_layout=True)
    for r, e in enumerate([False, True]):
        ax[r, 0].set_title(f"express_badness_with_uncertainty={e}")

        for c, f in enumerate([s.get_ok_data_for_wavelength, s.get_ok_data_for_time]):
            for i in range(3):
                x, y, sigma = f(i, y=quantity, express_badness_with_uncertainty=e)
                ax[r, c].scatter(x, y - i * 0.2, c=np.isfinite(sigma))



from ...imports import *

__all__ = ["fold", "mask_transit"]


def fold(self, period=None, t0=None, event="Mid-Transit"):
    """
    Fold this Rainbow to a period and reference epoch.

    Parameters
    ----------
    period : u.Quantity
        The orbital period of the planet (with astropy units of time).
    t0 : u.Quantity
        Any mid-transit epoch (with astropy units of time).
    event : str
        A description of the event that happens periodically.
        For example, you might want to switch this to
        'Mid-Eclipse' (as well as offsetting the `t0` by the
        appropriate amount relative to transit). This description
        may be used in plot labels.

    Returns
    -------
    folded : Rainbow
        The folded Rainbow
    """

    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("fold", locals())

    # warn
    if (period is None) or (t0 is None):
        message = """
        Folding to a transit period requires both
        `period` and `t0` be specified. Please try again.
        """
        cheerfully_suggest(message)
        return self

    # create a copy of the existing rainbow
    new = self._create_copy()

    # calculate predicted time from transit
    new.time = (((self.time - t0) + 0.5 * period) % period) - 0.5 * period
    # (the nudge by 0.5 period is to center on -period/2 to period/2)

    # change the default time label
    new.metadata["time_label"] = f"Time from {event}"

    # append the history entry to the new Rainbow
    new._record_history_entry(h)

    return new


def mask_transit(self, period, t0, duration, event="Mid-Transit"):
    """
    Mask

    Parameters
    ----------
    period : u.Quantity
        The orbital period of the planet (with astropy units of time).
    t0 : u.Quantity
        Any mid-transit epoch (with astropy units of time).
    duration : u.Quantity
        The total duration of the transit to remove.
    event : str
        A description of the event that happens periodically.
        For example, you might want to switch this to
        'Mid-Eclipse' (as well as offsetting the `t0` by the
        appropriate amount relative to transit). This description
        may be used in plot labels.

    Returns
    -------
    masked : Rainbow
        The Rainbow with the transit masked as not `ok`.
    """
    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("mask_transit", locals())

    if "fold" in self.history():
        raise ValueError(
            f"""
        It looks like this Rainbow has already been folded.
        Please run `.mask_transit` only on unfolded data.
        (It will refold while masking.)
        """
        )

    # fold and mask transit
    folded = self.fold(period=period, t0=t0, event=event)
    folded.timelike["ok"] = np.abs(folded.time) >= duration / 2

    # append the history entry to the new Rainbow
    folded._record_history_entry(h)

    return folded



"""
This setup.py file sets up our package to be installable on any computer,
so that folks can `import chromatic` from within any directory.

Thanks to this file, you can...

...tell python to look for code in the current directory (which you
can continue to edit), by typing *one* of the following commands:

`pip install -e .`
or
`python setup.py develop`

...move a copy of this code to your site-packages directory, where python will
be able to find it (but you won't be able to keep editing it), by typing *one*
of the following commands:

`pip install .`
or
`python setup.py install`

...upload the entire package to the Python Package Index, so that other folks
will be able to install your package via the simple `pip install chromatic-lightcurves`, by
running the following command:

`python setup.py release`

The template for this setup.py came was pieced together with help from
@christinahedges, @barentsen, @timothydmorton, and @dfm. Check them
out on github for more beautiful code!

[`python-packaging`](https://python-packaging.readthedocs.io/en/latest/index.html)
is a pretty useful resource too!
"""

# import our basic setup ingredients
from setuptools import setup, find_packages
import os, sys

# running `python setup.py release` from the command line will post to PyPI
if "release" in sys.argv[-1]:
    os.system("python setup.py sdist")
    # uncomment the next line to test out on test.pypi.com/project/tess-zap
    # os.system("twine upload --repository-url https://test.pypi.org/legacy/ dist/*")
    os.system("twine upload dist/*")
    os.system("rm -rf dist/chromatic-lightcurves*")
    sys.exit()

# a little kludge to get the version number from __version__
exec(open("chromatic/version.py").read())

# run the setup function
setup(
    # the name folks can use to search for this with pip
    name="chromatic-lightcurves",
    # what version of the code is this?
    version=__version__,
    # what's a short description of the package?
    description="Tools for working with spectroscopic light curves (of transiting exoplanets).",
    # what's a more detailed description?
    long_description="Read the complete documentation at https://zkbt.github.io/chromatic/",
    # who's the main author?
    author="Zach Berta-Thompson",
    # what's the main author's email?
    author_email="zach.bertathompson@colorado.edu",
    # what's the URL for the repository?
    url="https://github.com/zkbt/chromatic",
    # this figures out what subdirectories to include
    packages=find_packages(),
    # are there data that should be accessible when installed?
    include_package_data=True,
    # are there scripts to be copied into your $PATH?
    scripts=[],
    # some descriptions about this package (for searchability)
    classifiers=[
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Astronomy",
    ],
    # what other packages are needed? (must be pip-installable)
    install_requires=[
        "numpy",
        "scipy",
        "matplotlib>=3.5",
        "astropy>=4.0",
        "pandas",
        "tqdm",
        "batman-package>=2.4.9",
        "altair",
        "xarray",
        "h5py",
    ],
    # what version of Python is required?
    python_requires=">=3.7",  # f-strings are introduced in 3.6!
    # requirements in `key` will install with `pip install chromatic-lightcurves[key]`
    extras_require={
        "develop": [
            "pytest",
            "black",
            "jupyter",
            "ipython",
            "mkdocs",
            "mkdocs-material",
            "mkdocstrings",
            "pytkdocs[numpy-style]",
            "mkdocs-jupyter",
            "mkdocs-exclude",
            "twine",
            "pre-commit",
        ]
    },
    # (I think just leave this set to False)
    zip_safe=False,
    # under what license is this code released?
    license="MIT",
)



# import the general list of packages
from ...imports import *
from h5py import File

# define list of the only things that will show up in imports
__all__ = ["from_feinstein_numpy", "from_feinstein_h5"]


def from_feinstein_numpy(rainbow, filepath):
    """
    Populate a Rainbow from a file in the feinstein numpy format.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.
    filepath : str
        The path to the file to load.
    """

    time, wavelength, spectra, err = np.load(filepath, allow_pickle=True)

    rainbow.wavelike["wavelength"] = wavelength * u.micron * 1

    # populate a 1D array of times (with astropy units of time)
    times = time * u.day
    rainbow.timelike["time"] = times * 1

    # populate a 2D (row = wavelength, col = array of fluxes
    flux = np.zeros((len(wavelength), len(times)))
    uncertainty = np.zeros_like(flux)

    rainbow.fluxlike["flux"] = spectra.T * 1
    rainbow.fluxlike["uncertainty"] = err.T * 1


def from_feinstein_h5(self, filepath, order=1, version="opt"):
    """
    Populate a Rainbow from a file in the feinstein H5 format.

    Parameters
    ----------

    rainbow : Rainbow
        The object to be populated.
    filepath : str
        The path to the file to load.
    """

    f = File(filepath)

    astropy_times = Time(np.array(f["time"]), format="mjd", scale="tdb")
    self.set_times_from_astropy(astropy_times, is_barycentric=True)  # ???
    self.wavelike["wavelength"] = (
        np.array(f[f"wavelength_order_{order}"]) * u.micron * 1
    )
    for k in ["box_flux", "box_var", "opt_flux", "opt_var"]:
        self.fluxlike[k] = np.transpose(np.array(f[f"{k}_order_{order}"]) * 1)

    self.fluxlike["flux"] = self.fluxlike[f"{version}_flux"] * 1
    self.fluxlike["uncertainty"] = self.fluxlike[f"{version}_var"] * 1



from ...imports import *

__all__ = ["flag_outliers"]


def flag_outliers(
    self, sigma = 3
):
    """
    Flag outliers on a chromatic rainbow.

    Parameters
    ----------
    sigma : integer
        Standard deviations (sigmas) allowed for individual data points before they are flagged as outliers.

    Returns
    -------
    rainbow : Rainbow
        A new Rainbow object with the outliers flagged as 0's in the 'ok' array.
    """

    
    # create a history entry for this action (before other variables are defined)
    h = self._create_history_entry("flag_outliers", locals())
    
    # create a copy of the existing rainbow
    new = self._create_copy()
    
    list_of_rows = []
    for i, row in enumerate(new.flux):
        diffs = np.diff(row)
        mask = sigma_clip(diffs, sigma=sigma, maxiters=5, cenfunc = 'median')
        
        ls = np.hstack((np.asarray([True]),~mask.mask))
        new_ls = ls.copy()
        for i in range(1,len(ls)):
            if ((ls[i] == True) and (ls[i-1] == False)):
                new_ls[i-1] = True
                
        list_of_rows.append(new_ls)
            
    full_mask = np.array(list_of_rows)

    new.ok = new.ok * full_mask  
        
    # append the history entry to the new Rainbow
    new._record_history_entry(h)
    
    return new